/*
   Copyright (C) 2001,2002,2003,2004,2005,2006,2007 Keisuke Nishida
   Copyright (C) 2006,2007,2008,2009,2010 Roger While

   This file is part of OpenCOBOL.

   The OpenCOBOL compiler is free software: you can redistribute it
   and/or modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.

   OpenCOBOL is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with OpenCOBOL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "defaults.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <setjmp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef	HAVE_SIGNAL_H
#include <signal.h>
#endif

#ifdef	_WIN32
#define	WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <io.h>
#include <fcntl.h>
#endif

#ifdef	HAVE_KPATHSEA_GETOPT_H
#include <kpathsea/getopt.h>
#else
#ifdef	HAVE_GETOPT_H
#include <getopt.h>
#else
#ifndef	HAVE_CONFIG_H
#define	HAVE_CONFIG_H	1
#endif
#include "lib/getopt.h"
#endif
#endif

#ifdef	HAVE_LOCALE_H
#include <locale.h>
#endif

#include "tarstamp.h"

#include "cobc.h"
#include "tree.h"

/* Compile level */
#define	CB_LEVEL_PREPROCESS	1
#define	CB_LEVEL_TRANSLATE	2
#define	CB_LEVEL_COMPILE	3
#define	CB_LEVEL_ASSEMBLE	4
#define	CB_LEVEL_MODULE		5
#define	CB_LEVEL_LIBRARY	6
#define	CB_LEVEL_EXECUTABLE	7

#define	CB_IMSG_SIZE		24
#define	CB_IVAL_SIZE		(74 - CB_IMSG_SIZE - 4)

#define	COBC_ADD_STR(v,x,y,z)	v = cobc_add_str (v, &v##_size, x, y, z);

/* Global variables */

int	cb_source_format = CB_FORMAT_FIXED;	/* Default source format */

#ifdef	COB_EBCDIC_MACHINE
int	cb_display_sign = COB_DISPLAY_SIGN_EBCDIC;	/* 1 */
#else
int	cb_display_sign = COB_DISPLAY_SIGN_ASCII;	/* 0 */
#endif

#undef	COB_EXCEPTION
#define	COB_EXCEPTION(code,tag,name,critical) {name, 0x##code, 0},
struct cb_exception cb_exception_table[] = {
	{NULL, 0, 0},		/* CB_EC_ZERO */
#include "libcob/exception.def"
	{NULL, 0, 0}		/* CB_EC_MAX */
};
#undef	COB_EXCEPTION

#undef	CB_FLAG
#define	CB_FLAG(var,pdok,name,doc)	int var = 0;
#include "flag.def"
#undef	CB_FLAG

#undef	CB_WARNDEF
#define	CB_WARNDEF(var,name,wall,doc)	int var = 0;
#include "warning.def"
#undef	CB_WARNDEF

#ifdef	_MSC_VER
#define PATHSEPS ";"
#else
#define PATHSEPS ":"
#endif

int			cb_id = 0;
int			cb_attr_id = 0;
int			cb_literal_id = 0;
int			cb_field_id = 0;
int			cobc_flag_main = 0;
int			cb_flag_main = 0;
int			cobc_wants_debug = 0;

int			cb_stack_size = 0;
int			errorcount = 0;
int			warningcount = 0;
int			alt_ebcdic = 0;
int			exit_option = 0;
int			warningopt = 0;
int			no_physical_cancel = 0;

char			*cb_source_file = NULL;
char			*cb_oc_build_stamp = NULL;
/*
char			*source_name = NULL;
*/
const char		*demangle_name = NULL;
int			cb_source_line = 0;

FILE			*cb_storage_file = NULL;
const char		*cb_storage_file_name;

FILE			*cb_listing_file = NULL;
struct cb_text_list	*cb_include_list = NULL;
struct cb_text_list	*cb_extension_list = NULL;
struct cb_text_list	*cb_static_call_list = NULL;
struct cb_text_list	*cb_early_exit_list = NULL;
struct cb_define_struct	*cb_define_list = NULL;

int			cb_saveargc = 0;
char			**cb_saveargv = NULL;

const char		*cob_config_dir = NULL;

unsigned int		cobc_gen_listing = 0;

int			optimize_defs[COB_OPTIM_MAX] = { 0 };

/* Local variables */

static const char	*cob_cc;	/* C compiler */
static char		*cob_cflags;	/* C compiler flags */
static char		*cob_libs;	/* -l... */
static char		*cob_lib_paths;	/* -L... */
static char		*cob_include;	/* -I... */
static char		*cob_ldflags;	/* -Q / COB_LDFLAGS */
static const char	*cob_copy_dir;
static const char	*cob_tmpdir;	/* temporary directory */

static size_t		cob_cflags_size;
static size_t		cob_libs_size;
static size_t		cob_lib_paths_size;
static size_t		cob_include_size;
static size_t		cob_ldflags_size;

static char		*save_temps_dir;

static char		*cobc_list_dir;
static char		*cobc_list_file;

static char		*output_name;
static char		*cobc_buffer;
static char		*cobc_temp_buffer;
static char		*output_name_buff;
static char		*basename_buffer;
static size_t		basename_len;
static size_t		cobc_buffer_size;

static struct filename	*file_list;

#if	0	/* Deprecated */
static const char	*cb_depend_target = NULL;
FILE			*cb_depend_file = NULL;
struct cb_text_list	*cb_depend_list = NULL;
#endif

static size_t		iparams = 0;

static unsigned int	cb_compile_level = 0;
static unsigned int	local_level = 0;

static int		iargs;

static int		wants_nonfinal = 0;
static int		cobc_flag_module = 0;
static int		cobc_flag_library = 0;
static int		save_temps = 0;
static int		save_csrc = 0;
static int		verbose_output = 0;
static int		cob_iteration = 0;
#ifndef _WIN32
static pid_t		cob_process_id = 0;
#endif

#ifdef	_MSC_VER
#if	_MSC_VER >= 1400
static const char	*manicmd;
#endif
static const char	*manilink;
#endif

static int		strip_output = 0;
static int		gflag_set = 0;

static jmp_buf		cobc_jmpbuf;

static const char	*const cob_csyns[] = {
#ifndef	COB_EBCDIC_MACHINE
	"NULL",
	"P_cancel",
	"P_initialize",
	"P_ret_initialize",
	"P_switch",
#endif
	"_Bool",
	"_Complex",
	"_Imaginary",
	"alignof",
	"asm",
	"auto",
	"bool",
	"break",
	"case",
	"catch",
	"char",
	"class",
	"const",
	"const_cast",
	"continue",
	"default",
	"delete",
	"do",
	"double",
	"dynamic_cast",
	"else",
	"enum",
	"exit_program",
	"explicit",
	"extern",
	"false",
	"float",
	"for",
	"frame_pointer",
	"frame_stack",
	"friend",
	"goto",
	"if",
	"inline",
	"int",
	"long",
	"mutable",
	"namespace",
	"new",
	"offsetof",
	"operator",
	"private",
	"protected",
	"public",
	"register",
	"reinterpret_cast",
	"restrict",
	"return",
	"short",
	"signed",
	"sizeof",
	"static",
	"static_cast",
	"struct",
	"switch",
	"template",
	"this",
	"throw",
	"true",
	"try",
	"typedef",
	"typeid",
	"typename",
	"typeof",
	"union",
	"unsigned",
	"using",
	"virtual",
	"void",
	"volatile",
#ifndef	COB_EBCDIC_MACHINE
	"wchar_t"
#else
	"wchar_t",
	"NULL",
	"P_cancel",
	"P_initialize",
	"P_ret_initialize",
	"P_switch"
#endif
};

#define COB_NUM_CSYNS	sizeof(cob_csyns) / sizeof(char *)

#ifdef	HAVE_SIGNAL_H
#ifdef	HAVE_SIGACTION
typedef struct sigaction	cob_sighandler_t;
#else
typedef void (*cob_sighandler_t) (int);
#endif
#ifdef	SIGHUP
static cob_sighandler_t		hupsig;
#endif
#ifdef	SIGINT
static cob_sighandler_t		intsig;
#endif
#ifdef	SIGQUIT
static cob_sighandler_t		qutsig;
#endif
#endif

static const char short_options[] = "hVivECScbmxOPgwo:I:L:l:D:K:k:";

#define	CB_NO_ARG	no_argument
#define	CB_RQ_ARG	required_argument
#define	CB_OP_ARG	optional_argument

static const struct option long_options[] = {
	{"help",		CB_NO_ARG, NULL, 'h'},
	{"version",		CB_NO_ARG, NULL, 'V'},
/* getopt_long_only has a problem with eg. -xv - remove
	{"verbose",		CB_NO_ARG, NULL, 'v'},
*/
	{"info",		CB_NO_ARG, NULL, 'i'},
	{"list-reserved",	CB_NO_ARG, NULL, '5'},
	{"list-intrinsics",	CB_NO_ARG, NULL, '6'},
	{"list-mnemonics",	CB_NO_ARG, NULL, '7'},
	{"list-system",		CB_NO_ARG, NULL, '8'},
	{"save-temps",		CB_OP_ARG, NULL, '_'},
	{"std",			CB_RQ_ARG, NULL, '$'},
	{"conf",		CB_RQ_ARG, NULL, '&'},
	{"debug",		CB_NO_ARG, NULL, 'd'},
	{"ext",			CB_RQ_ARG, NULL, 'e'},
	{"stack-size",		CB_RQ_ARG, NULL, 'M'},
	{"free",		CB_NO_ARG, &cb_source_format, CB_FORMAT_FREE},
	{"fixed",		CB_NO_ARG, &cb_source_format, CB_FORMAT_FIXED},
	{"static",		CB_NO_ARG, &cb_flag_static_call, 1},
	{"dynamic",		CB_NO_ARG, &cb_flag_static_call, 0},
	{"O2",			CB_NO_ARG, NULL, '2'},
	{"Os",			CB_NO_ARG, NULL, 's'},
	{"Q",			CB_RQ_ARG, NULL, 'Q'},
	{"A",			CB_RQ_ARG, NULL, 'A'},
	{"P",			CB_OP_ARG, NULL, 'P'},
	{"Xref",		CB_NO_ARG, NULL, 'X'},
#if	0	/* Deprecated */
	{"MT",			CB_RQ_ARG, NULL, '%'},
	{"MF",			CB_RQ_ARG, NULL, '@'},
#endif
	{"Wall",		CB_NO_ARG, NULL, 'W'},
	{"W",			CB_NO_ARG, NULL, 'Z'},

#undef	CB_FLAG
#define	CB_FLAG(var,pdok,name,doc)			\
	{"f"name,		CB_NO_ARG, &var, 1},	\
	{"fno-"name,		CB_NO_ARG, &var, 0},
#include "flag.def"
#undef	CB_FLAG

#undef	CB_WARNDEF
#define	CB_WARNDEF(var,name,wall,doc)			\
	{"W"name,		CB_NO_ARG, &var, 1},	\
	{"Wno-"name,		CB_NO_ARG, &var, 0},
#include "warning.def"
#undef	CB_WARNDEF

	{NULL,			0, NULL, 0}
};

#undef	CB_ARG_NO
#undef	CB_ARG_RQ
#undef	CB_ARG_OP

/* cobc functions */

/*
 * Global functions
 */

void
cobc_too_many_errors (void)
{
	fprintf (stderr, _("Too many errors - Aborting compilation\n"));
	(void)longjmp (cobc_jmpbuf, 1);
}

void
cobc_abort (const char *filename, const int linenum)
{
	fprintf (stderr, _("%s:%d: Internal compiler error\n"), filename, linenum);
	(void)longjmp (cobc_jmpbuf, 1);
}

void
cobc_tree_cast_error (cb_tree x, const char *filen, const int linenum,
		      const int tagnum)
{
	fprintf (stderr, _("%s:%d: Invalid type cast from '%s'\n"),
		filen, linenum, x ? cb_name (x) : "null");
	fprintf (stderr, "Tag 1 %d Tag 2 %d\n", x ? (int)CB_TREE_TAG(x) : 0,
		tagnum);
	(void)longjmp (cobc_jmpbuf, 1);
}

#if	!defined(__GNUC__) && defined(COB_TREE_DEBUG)
cb_tree
cobc_tree_cast_check (const cb_tree x, const int tag, const char *file,
		      const int line)
{
	if (!x || x == cb_error_node || CB_TREE_TAG (x) != tag) {
		cobc_tree_cast_error (x, file, line, tag);
	}
	return x;
}
#endif

void *
cobc_malloc (const size_t size)
{
	void *mptr;

	mptr = calloc (1, size);
	if (unlikely(!mptr)) {
		fprintf (stderr, _("Cannot allocate %d bytes of memory - Aborting\n"),
			 (int)size);
		(void)longjmp (cobc_jmpbuf, 1);
	}
	return mptr;
}

void *
cobc_strdup (const char *dupstr)
{
	void	*p;
	size_t	n;

	if (unlikely(!dupstr)) {
		fprintf (stderr, _("Call to cobc/strdup with NULL pointer\n"));
		(void)longjmp (cobc_jmpbuf, 1);
	}
	n = strlen (dupstr);
	p = cobc_malloc (n + 1);
	memcpy (p, dupstr, n);
	return p;
}

static char *
cobc_stradd_dup (const char *str1, const char *str2)
{
	char	*p;
	size_t	m;
	size_t	n;

	if (unlikely(!str1 || !str2)) {
		fprintf (stderr, _("Call to cobc_stradd_dup with NULL pointer\n"));
		(void)longjmp (cobc_jmpbuf, 1);
	}
	m = strlen (str1);
	n = strlen (str2);
	p = cobc_malloc (m + n + 1);
	memcpy (p, str1, m);
	memcpy (p + m, str2, n);
	return p;
}

void *
cobc_realloc (void *prevptr, const size_t size)
{
	void *mptr;

	mptr = realloc (prevptr, size);
	if (unlikely(!mptr)) {
		fprintf (stderr,
			 _("Cannot reallocate %d bytes of memory - Aborting\n"),
			 (int)size);
		fflush (stderr);
		(void)longjmp (cobc_jmpbuf, 1);
	}
	return mptr;
}

struct cb_text_list *
cb_text_list_add (struct cb_text_list *list, const char *text)
{
	struct cb_text_list	*p;
	struct cb_text_list	*l;

	p = cobc_malloc (sizeof (struct cb_text_list));
	p->text = cobc_strdup (text);
	p->next = NULL;
	if (!list) {
		return p;
	} else {
		for (l = list; l->next; l = l->next) {
			;
		}
		l->next = p;
		return list;
	}
}

unsigned int
cb_set_value (struct cb_define_struct *p, const char *value)
{
	size_t		size;
	const char	*s;
	unsigned int	dotseen;
	int		sign;
	int		int_part;
	int		dec_part;

	if (!value) {
		p->deftype = PLEX_DEF_NONE;
		p->value = NULL;
		p->sign = 0;
		p->int_part = 0;
		p->dec_part = 0;
		return 0;
	}

	if (*value == '"' || *value == '\'') {
		sign = *value;
		p->value = cobc_strdup (value + 1);
		size = strlen (p->value) - 1;
		if (sign != p->value[size]) {
			free (p->value);
			p->value = NULL;
			p->deftype = PLEX_DEF_NONE;
			return 1;
		}
		p->value[size] = 0;
		p->deftype = PLEX_DEF_LIT;
		p->sign = 0;
		p->int_part = 0;
		p->dec_part = 0;
		return 0;
	}

	p->value = cobc_strdup (value);
	p->deftype = PLEX_DEF_NUM;
	p->sign = 0;
	p->int_part = 0;
	p->dec_part = 0;

	sign = 0;
	if (*value == '+') {
		value++;
	} else if (*value == '-') {
		value++;
		sign = 1;
	}
	int_part = 0;
	dec_part = 0;
	size = 0;
	dotseen = 0;
	s = value;
	for ( ; *s; ++s, ++size) {
		if (*s == '.') {
			if (dotseen) {
				p->deftype = PLEX_DEF_NONE;
				return 1;
			}
			dotseen = 1;
			continue;
		}
		if (*s > '9' || *s < '0') {
			p->deftype = PLEX_DEF_NONE;
			return 1;
		}
		if (!dotseen) {
			int_part = (int_part * 10) + (*s - '0');
		} else {
			dec_part = (dec_part * 10) + (*s - '0');
		}
	}

	if (!int_part && !dec_part) {
		sign = 0;
	}
	p->sign = sign;
	p->int_part = int_part;
	p->dec_part = dec_part;
	return 0;
}

static struct cb_define_struct *
cb_define_list_add (struct cb_define_struct *list, const char *text)
{
	struct cb_define_struct	*p;
	struct cb_define_struct	*l;
	char			*s;
	char			*x;

	x = cobc_strdup (text);
	s = strtok (x, "=");

	/* Check duplicate */
	for (l = list; l; l = l->next) {
		if (!strcasecmp (s, l->name)) {
			fprintf (stderr, _("Duplicate define '%s' - Ignoring\n"), s);
			fflush (stderr);
			free (x);
			return list;
		}
	}

	p = cobc_malloc (sizeof (struct cb_define_struct));
	p->next = NULL;
	p->name = cobc_strdup (s);
	p->deftype = PLEX_DEF_NONE;
	s = strtok (NULL, "");
	if (cb_set_value (p, s)) {
		free (p->name);
		free (p);
		free (x);
		return NULL;
	}

	free (x);

	if (!list) {
		return p;
	} else {
		for (l = list; l->next; l = l->next) {
			;
		}
		l->next = p;
		return list;
	}
}

static int
cobc_bcompare (const void *p1, const void *p2)
{
	const void	**tptr;

	tptr = (const void **)p2;
	return strcmp (p1, *tptr);
}

size_t
cobc_check_valid_name (const char *name, const unsigned int prechk)
{
	size_t	len;

	len = strlen (name);
	if (len < 1 || len > 31) {
		return 1;
	}
	if (*name == '_' || *name == ' ') {
		return 1;
	}
	if (prechk && len > 3 &&
	    (!memcmp (name, "cob_", 4) || !memcmp (name, "COB_", 4))) {
		return 1;
	}
	if (bsearch (name, cob_csyns, COB_NUM_CSYNS,
		     sizeof (char *), cobc_bcompare)) {
		return 1;
	}
	return 0;
}

/*
 * Local functions
 */

static char *
cobc_getenv (const char *env)
{
	char	*p;

	p = getenv (env);
	if (!p) {
		return NULL;
	}
	if (*p == 0 || *p == ' ') {
		return NULL;
	}
	return cobc_strdup (p);
}

static char *
cobc_add_str (char *var, size_t *cursize, const char *s1, const char *s2,
	      const char *s3)
{
	char	*p;
	size_t	calcsize;
	size_t	origsize;

	calcsize = strlen (var);
	origsize = calcsize;
	if (s1) {
		calcsize += strlen (s1);
	}
	if (s2) {
		calcsize += strlen (s2);
	}
	if (s3) {
		calcsize += strlen (s3);
	}
	if (calcsize >= 131072) {
		/* Arbitrary limit */
		fprintf (stderr, _("Parameter buffer size exceeded\n"));
		exit (1);
	}
	if (calcsize >= *cursize) {
		while (*cursize <= calcsize) {
			*cursize *= 2;
		}
		p = cobc_malloc (*cursize);
		memcpy (p, var, origsize);
		free (var);
	} else {
		p = var;
	}
	if (s1) {
		strcat (p, s1);
	}
	if (s2) {
		strcat (p, s2);
	}
	if (s3) {
		strcat (p, s3);
	}
	return p;
}

static void
cobc_check_action (const char *name)
{
	struct stat		st;

	if (name && !stat (name, &st)) {
		if (!save_temps) {
			unlink (name);
		} else if (save_temps_dir) {
			if (!cobc_temp_buffer) {
				cobc_temp_buffer = cobc_malloc (COB_MEDIUM_BUFF);
			}
			snprintf (cobc_temp_buffer, COB_MEDIUM_MAX, "%s/%s",
				  save_temps_dir, name);
			/* Remove possible target file - ignore return */
			(void)unlink (cobc_temp_buffer);
			if (rename (name, cobc_temp_buffer)) {
				fprintf (stderr, "Warning - Could not move temporary file to %s\n", cobc_buffer);
			}
		}
	}
}

static void COB_NOINLINE
cobc_clean_up (const unsigned int status)
{
	struct filename		*fn;
	struct local_filename	*lf;
	int			i;
	struct stat		st;

	if (cb_listing_file) {
		fclose (cb_listing_file);
		cb_listing_file = NULL;
	}
	if (ppin) {
		fclose (ppin);
		ppin = NULL;
	}

	if (ppout) {
		fclose (ppout);
		ppout = NULL;
	}
	if (yyin) {
		fclose (yyin);
		yyin = NULL;
	}

	if (yyout) {
		fclose (yyout);
		yyout = NULL;
	}

	for (fn = file_list; fn; fn = fn->next) {
		if (fn->need_assemble &&
		    (status || cb_compile_level > CB_LEVEL_ASSEMBLE ||
		    (cb_compile_level == CB_LEVEL_ASSEMBLE && save_temps))) {
			cobc_check_action (fn->object);
		}
		if (save_csrc) {
			continue;
		}
		if (fn->need_preprocess &&
		    (status || cb_compile_level > CB_LEVEL_PREPROCESS ||
		    (cb_compile_level == CB_LEVEL_PREPROCESS && save_temps))) {
			cobc_check_action (fn->preprocess);
		}
		if (fn->need_translate &&
		    (status || cb_compile_level > CB_LEVEL_TRANSLATE ||
		    (cb_compile_level == CB_LEVEL_TRANSLATE && save_temps))) {
			cobc_check_action (fn->translate);
			cobc_check_action (fn->trstorage);
			if (fn->localfile) {
				for (lf = fn->localfile; lf; lf = lf->next) {
					cobc_check_action (lf->local_name);
				}
			} else if (fn->translate) {
				/* If we get syntax errors, we do not
				   know the number of local include files */
				sprintf (cobc_buffer, "%s.l.h", fn->translate);
				for (i = 0; i < 30; ++i) {
					if (i) {
						sprintf (cobc_buffer, "%s.l%d.h",
							fn->translate, i);
					}
					if (!stat (cobc_buffer, &st)) {
						unlink (cobc_buffer);
					} else {
						break;
					}
				}
			}
		}
	}
}

static void
cobc_terminate (const char *str)
{
	fprintf (stderr, "cobc: ");
	fflush (stderr);
	perror (str);
	fflush (stderr);
	cobc_clean_up (1);
	exit (1);
}

#ifdef	HAVE_SIGNAL_H
static void
cobc_sig_handler (int sig)
{
	save_temps = 0;
	cobc_clean_up (1);
	switch (sig) {
#ifdef	SIGHUP
	case SIGHUP:
#ifdef	HAVE_SIGACTION
		if ((hupsig.sa_handler != SIG_IGN) &&
		    (hupsig.sa_handler != SIG_DFL)) {
			hupsig.sa_handler (SIGHUP);
#else
		if ((hupsig != SIG_IGN) && (hupsig != SIG_DFL)) {
			(*hupsig) (SIGHUP);
#endif
		}
		break;
#endif
#ifdef	SIGINT
	case SIGINT:
#ifdef	HAVE_SIGACTION
		if ((intsig.sa_handler != SIG_IGN) &&
		    (intsig.sa_handler != SIG_DFL)) {
			intsig.sa_handler (SIGINT);
#else
		if ((intsig != SIG_IGN) && (intsig != SIG_DFL)) {
			(*intsig) (SIGINT);
#endif
		}
		break;
#endif
#ifdef	SIGQUIT
	case SIGQUIT:
#ifdef	HAVE_SIGACTION
		if ((qutsig.sa_handler != SIG_IGN) &&
		    (qutsig.sa_handler != SIG_DFL)) {
			qutsig.sa_handler (SIGQUIT);
#else
		if ((qutsig != SIG_IGN) && (qutsig != SIG_DFL)) {
			(*qutsig) (SIGQUIT);
#endif
		}
		break;
#endif
	}
	exit (sig);
}
#endif

/*
 * Command line
 */

static void COB_NOINLINE
cobc_print_version (void)
{
	printf ("cobc (%s) %s.%d\n",
		PACKAGE_NAME, PACKAGE_VERSION, PATCH_LEVEL);
	puts ("Copyright (C) 2001,2002,2003,2004,2005,2006,2007 Keisuke Nishida");
	puts ("Copyright (C) 2006,2007,2008,2009,2010 Roger While");
	puts ("This is free software; see the source for copying conditions.  There is NO");
	puts ("warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.");
	printf ("Built     %s\n", cb_oc_build_stamp);
	printf ("Packaged  %s\n", COB_TAR_DATE);
#ifdef	__xlc__
	printf ("C version %s\n", COBXSTRINGIFY(__xlc__));
#elif	defined(__VERSION__)
	printf ("C version %s\n", COBXSTRINGIFY(__VERSION__));
#endif
}

static void COB_NOINLINE
cobc_cmd_print (const char *cmd)
{
	char	*p;
	char	*token;
	size_t	n;
	size_t	toklen;

	/* Check if it fits in 80 characters */
	if (strlen (cmd) < 64) {
		fprintf (stderr, "Executing:\t%s\n", (char *)cmd);
		fflush (stderr);
		return;
	}
	fprintf (stderr, "Executing:\t");
	p = cobc_strdup (cmd);
	n = 0;
	token = strtok (p, " ");
	for (; token; token = strtok (NULL, " ")) {
		toklen = strlen (token) + 1;
		if ((n + toklen) > 63) {
			fprintf(stderr, "\n\t\t");
			n = 0;
		}
		fprintf (stderr, "%s%s", (n ? " " : ""), token);
		n += toklen;
	}
	free (p);
	fprintf (stderr, "\n");
	fflush (stderr);
}

static void COB_NOINLINE
cobc_var_print (const char *msg, const char *val)
{
	char	*p;
	char	*token;
	size_t	n;
	size_t	toklen;

	printf ("%-*.*s : ", CB_IMSG_SIZE, CB_IMSG_SIZE, msg);
	if (strlen(val) <= CB_IVAL_SIZE) {
		printf("%s\n", val);
		return;
	}
	p = cobc_strdup (val);
	n = 0;
	token = strtok (p, " ");
	for (; token; token = strtok (NULL, " ")) {
		toklen = strlen (token) + 1;
		if ((n + toklen) > CB_IVAL_SIZE) {
			if (n) {
				printf ("\n%*.*s", CB_IMSG_SIZE + 3,
					CB_IMSG_SIZE + 3, " ");
			}
			n = 0;
		}
		printf ("%s%s", (n ? " " : ""), token);
		n += toklen;
	}
	printf ("\n");
	free (p);
}

static void COB_NOINLINE
cobc_print_info (void)
{
	char	buff[4];

	cobc_print_version ();
	printf ("Build information\n");
	cobc_var_print ("Build environment",	COB_BLD_BUILD);
	cobc_var_print ("CC",			COB_BLD_CC);
	cobc_var_print ("CPPFLAGS",		COB_BLD_CPPFLAGS);
	cobc_var_print ("CFLAGS",		COB_BLD_CFLAGS);
	cobc_var_print ("LD",			COB_BLD_LD);
	cobc_var_print ("LDFLAGS",		COB_BLD_LDFLAGS);
	printf ("\nOpenCOBOL information\n");
	cobc_var_print ("COB_CC",		COB_CC);
	cobc_var_print ("COB_CFLAGS",		COB_CFLAGS);
	cobc_var_print ("COB_LDFLAGS",		COB_LDFLAGS);
	cobc_var_print ("COB_LIBS",		COB_LIBS);
	cobc_var_print ("COB_CONFIG_DIR",	COB_CONFIG_DIR);
	cobc_var_print ("COB_COPY_DIR",		COB_COPY_DIR);
	cobc_var_print ("COB_LIBRARY_PATH",	COB_LIBRARY_PATH);
	cobc_var_print ("COB_MODULE_EXT",	COB_MODULE_EXT);
	cobc_var_print ("COB_EXEEXT",		COB_EXEEXT);

#if	defined(USE_LIBDL) || defined(_WIN32)
	cobc_var_print ("Dynamic loading",	"System");
#else
	cobc_var_print ("Dynamic loading",	"Libtool");
#endif

#ifdef	COB_PARAM_CHECK
	cobc_var_print ("\"CBL_\" param check",	"Enabled");
#else
	cobc_var_print ("\"CBL_\" param check",	"Disabled");
#endif

	sprintf (buff, "%d", WITH_VARSEQ);
	cobc_var_print ("Variable format",	buff);

#ifdef	COB_LI_IS_LL
	cobc_var_print ("BINARY-C-LONG",	"8 bytes");
#else
	cobc_var_print ("BINARY-C-LONG",	"4 bytes");
#endif

#ifdef	WITH_SEQRA_EXTFH
	cobc_var_print ("Sequential handler",	"External");
#else
	cobc_var_print ("Sequential handler",	"Internal");
#endif
#ifdef	WITH_INDEX_EXTFH
	cobc_var_print ("ISAM handler",		"External");
#endif
#ifdef	WITH_DB
	cobc_var_print ("ISAM handler",		"BDB");
#endif
#ifdef	WITH_CISAM
	cobc_var_print ("ISAM handler",		"C-ISAM (Experimental)");
#endif
#ifdef	WITH_DISAM
	cobc_var_print ("ISAM handler",		"D-ISAM (Experimental)");
#endif
#ifdef	WITH_VBISAM
	cobc_var_print ("ISAM handler",		"VBISAM (Experimental)");
#endif
}

static void COB_NOINLINE
cobc_print_usage (void)
{
	printf ("Usage: cobc [options] file...\n\n");
	puts (_("Options:"));
	puts (_("  -help                 Display this message"));
	puts (_("  -version, -V          Display compiler version"));
	puts (_("  -info, -i             Display compiler build information"));
	puts (_("  -v                    Display the commands invoked by the compiler"));
	puts (_("  -x                    Build an executable program"));
	puts (_("  -m                    Build a dynamically loadable module (default)"));
	puts (_("  -std=<dialect>        Warnings/features for a specific dialect :"));
	puts (_("                          cobol2002   Cobol 2002"));
	puts (_("                          cobol85     Cobol 85"));
	puts (_("                          ibm         IBM Compatible"));
	puts (_("                          mvs         MVS Compatible"));
	puts (_("                          bs2000      BS2000 Compatible"));
	puts (_("                          mf          Micro Focus Compatible"));
	puts (_("                          default     When not specified"));
	puts (_("                        See config/default.conf and config/*.conf"));
	puts (_("  -free                 Use free source format"));
	puts (_("  -fixed                Use fixed source format (default)"));
	puts (_("  -O, -O2, -Os          Enable optimization"));
	puts (_("  -g                    Enable C compiler debug / stack check / trace"));
	puts (_("  -debug                Enable all run-time error checking"));
	puts (_("  -o <file>             Place the output into <file>"));
	puts (_("  -b                    Combine all input files into a single"));
	puts (_("                        dynamically loadable module"));
	puts (_("  -E                    Preprocess only; do not compile, assemble or link"));
	puts (_("  -C                    Translation only; convert COBOL to C"));
	puts (_("  -S                    Compile only; output assembly file"));
	puts (_("  -c                    Compile and assemble, but do not link"));
	puts (_("  -P(=<dir or file>)    Generate preprocessed program listing (.lst)"));
	puts (_("  -Xref                 Generate cross reference through 'cobxref'"));
	puts (_("                        (V. Coen's 'cobxref' must be in path)"));
	puts (_("  -I <directory>        Add <directory> to copy/include search path"));
	puts (_("  -L <directory>        Add <directory> to library search path"));
	puts (_("  -l <lib>              Link the library <lib>"));
	puts (_("  -A <options>          Add <options> to the C compile phase"));
	puts (_("  -Q <options>          Add <options> to the C link phase"));
	puts (_("  -D <define>           Pass <define> to the C compiler"));
	puts (_("  -K <entry>            Generate CALL to <entry> as static"));
	puts (_("  -conf=<file>          User defined dialect configuration - See -std="));
	puts (_("  -list-reserved        Display reserved words"));
	puts (_("  -list-intrinsics      Display intrinsic functions"));
	puts (_("  -list-mnemonics       Display mnemonic names"));
	puts (_("  -list-system          Display system routines"));
	puts (_("  -save-temps(=<dir>)   Save intermediate files"));
	puts (_("                        - (default current directory)"));
#if	0	/* Deprecated */
	puts (_("  -MT <target>          Set target file used in dependency list"));
	puts (_("  -MF <file>            Place dependency list into <file>"));
#endif
	puts (_("  -ext <extension>      Add default file extension"));
	puts (_("  -stack-size=n         Set perform stack size"));
	puts (_("                        - (>= 64, <= 512, default 255)"));
	puts ("");
	puts (_("  -W                    Enable ALL warnings"));
	puts (_("  -Wall                 Enable all warnings except as noted below"));
#undef	CB_WARNDEF
#define	CB_WARNDEF(var,name,wall,doc)		\
	printf ("  -W%-19s %s", name, doc); \
	if (!wall) { \
		puts (_("\n\t\t\t- (NOT set with -Wall)")); \
	} else { \
		printf ("\n"); \
	}
#include "warning.def"
#undef	CB_WARNDEF
	puts ("");
#undef	CB_FLAG
#define	CB_FLAG(var,pdok,name,doc)	\
	if (pdok)			\
		printf ("  -f%-19s %s\n", name, doc);
#include "flag.def"
#undef	CB_FLAG
	puts ("");
}

static void
cobc_options_error_nonfinal (void)
{
	fprintf (stderr, _("Only one of options 'E', 'S', 'C', 'c' may be specified\n"));
	exit (1);
}

static void
cobc_options_error_build (void)
{
	fprintf (stderr, _("Only one of options 'm', 'x', 'b' may be specified\n"));
	exit (1);
}

static int
process_command_line (int argc, char *argv[])
{
	struct cb_define_struct	*p;
	size_t			osize;
	int			c;
	int			idx;
	int			n;
	enum cob_exception_id	i;
	struct stat		st;
	char			ext[COB_MINI_BUFF];

	/* Enable default I/O exceptions */
	CB_EXCEPTION_ENABLE (COB_EC_I_O) = 1;

	while ((c = getopt_long_only (argc, argv, short_options, long_options, &idx)) >= 0) {
		switch (c) {
		case 0:
			/* EOF */
			break;

		case '?':
			/* Unknown option */
			exit (1);

		case 'h':
			/* --help */
			cobc_print_usage ();
			exit (0);

		case 'V':
			/* --version */
			cobc_print_version ();
			exit_option = 1;
			break;

		case 'i':
			/* --info */
			cobc_print_info ();
			exit_option = 1;
			break;

		case '5':
			/* --list-reserved */
			cb_list_reserved ();
			exit_option = 1;
			break;

		case '6':
			/* --list-intrinsics */
			cb_list_intrinsics ();
			exit_option = 1;
			break;

		case '7':
			/* --list-mnemonics */
			cb_list_mnemonics ();
			exit_option = 1;
			break;

		case '8':
			/* --list-system */
			cb_list_system ();
			exit_option = 1;
			break;

		case 'E':
			/* -E : Preprocess */
			if (wants_nonfinal) {
				cobc_options_error_nonfinal ();
			}
			wants_nonfinal = 1;
			cb_compile_level = CB_LEVEL_PREPROCESS;
			break;

		case 'C':
			/* -C : Generate C code */
			if (wants_nonfinal) {
				cobc_options_error_nonfinal ();
			}
			wants_nonfinal = 1;
			cb_compile_level = CB_LEVEL_TRANSLATE;
			break;

		case 'S':
			/* -S : Generate assembler code */
			if (wants_nonfinal) {
				cobc_options_error_nonfinal ();
			}
			wants_nonfinal = 1;
			cb_compile_level = CB_LEVEL_COMPILE;
			break;

		case 'c':
			/* -c : Generate C object code */
			if (wants_nonfinal) {
				cobc_options_error_nonfinal ();
			}
			wants_nonfinal = 1;
			cb_compile_level = CB_LEVEL_ASSEMBLE;
			break;

		case 'b':
			/* -b : Generate combined library module */
			if (cobc_flag_main || cobc_flag_module) {
				cobc_options_error_build ();
			}
			cobc_flag_library = 1;
			no_physical_cancel = 1;
			break;

		case 'm':
			/* -m : Generate loadable module (default) */
			if (cobc_flag_main || cobc_flag_library) {
				cobc_options_error_build ();
			}
			cobc_flag_module = 1;
			break;

		case 'x':
			/* -x : Generate executable */
			if (cobc_flag_module || cobc_flag_library) {
				cobc_options_error_build ();
			}
			cobc_flag_main = 1;
			cb_flag_main = 1;
			no_physical_cancel = 1;
			break;

		case 'v':
			/* -v : Verbose reporting */
			verbose_output = 1;
			break;

		case 'o':
			/* -o : Output file */
			osize = strlen (optarg);
			if (osize > COB_SMALL_MAX) {
				fprintf (stderr, _("Invalid output file name\n"));
				exit (1);
			}
			output_name = cobc_strdup (optarg);
			/* Allocate buffer plus extension reserve */
			output_name_buff = cobc_malloc (osize + 32);
			break;

		case 'O':
			/* -O : Optimize */
#ifdef	_MSC_VER
			COBC_ADD_STR (cob_cflags, " /O1", NULL, NULL);
#elif   defined(__SUNPRO_C)
			COBC_ADD_STR (cob_cflags, " -xO1", NULL, NULL);
#else
			COBC_ADD_STR (cob_cflags, " -O", NULL, NULL);
#endif
			break;

		case '2':
			/* -O2 : Optimize */
			strip_output = 1;
#ifdef	_MSC_VER
			COBC_ADD_STR (cob_cflags, " /O2", NULL, NULL);
#elif defined(__hpux) && !defined(__GNUC__)
			COBC_ADD_STR (cob_cflags, " +O2", NULL, NULL);
#elif   defined(__SUNPRO_C)
			COBC_ADD_STR (cob_cflags, " -xO2", NULL, NULL);
#else
			COBC_ADD_STR (cob_cflags, " -O2", NULL, NULL);
#endif
			break;

		case 's':
			/* -Os : Optimize */
			strip_output = 1;
#ifdef	_MSC_VER
			COBC_ADD_STR (cob_cflags, " /Os", NULL, NULL);
#elif defined(__hpux) && !defined(__GNUC__)
			COBC_ADD_STR (cob_cflags, " +O2 +Osize", NULL, NULL);
#elif   defined(__SUNPRO_C)
			COBC_ADD_STR (cob_cflags, " -xO1 -xspace", NULL, NULL);
#else
			COBC_ADD_STR (cob_cflags, " -Os", NULL, NULL);
#endif
			break;

		case 'g':
			/* -g : Generate C debug code */
			save_csrc = 1;
			gflag_set = 1;
			cb_flag_stack_check = 1;
			cb_flag_source_location = 1;
#ifndef	_MSC_VER
			COBC_ADD_STR (cob_cflags, " -g", NULL, NULL);
#endif
			break;

		case '$':
			/* -std=<xx> : Specify dialect */
			snprintf (ext, COB_MINI_MAX, "%s.conf", optarg);
			if (cb_load_std (ext) != 0) {
				fprintf (stderr, _("Invalid option -std=%s\n"), optarg);
				exit (1);
			}
			break;

		case '&':
			/* -conf=<xx> : Specify dialect configuration file */
			if (strlen (optarg) > COB_SMALL_MAX) {
				fprintf (stderr, _("Invalid -conf parameter\n"));
				exit (1);
			}
			if (cb_load_conf (optarg, 1, 0) != 0) {
				exit (1);
			}
			break;

		case 'd':
			/* -debug : Turn on OC debugging */
			/* Turn on all exception conditions */
			for (i = (enum cob_exception_id)1; i < COB_EC_MAX; ++i) {
				CB_EXCEPTION_ENABLE (i) = 1;
			}
			cb_flag_source_location = 1;
			cb_flag_trace = 1;
			cb_flag_stack_check = 1;
			cobc_wants_debug = 1;
			break;

		case '_':
			/* --save-temps : Save intermediary files */
			save_temps = 1;
			if (optarg) {
				if (stat (optarg, &st) != 0 ||
				    !(S_ISDIR (st.st_mode))) {
					fprintf (stderr, _("Warning - '%s' is not a directory, defaulting to current directory\n"), optarg);
					fflush (stderr);
				} else {
					save_temps_dir = cobc_strdup (optarg);
				}
			}
			break;

		case 'P':
			/* -P : Generate listing */
			if (optarg) {
				if (!stat (optarg, &st) && S_ISDIR (st.st_mode)) {
					cobc_list_dir = cobc_strdup (optarg);
				} else {
					cobc_list_file = cobc_strdup (optarg);
				}
			}
			if (!cobc_gen_listing) {
				cobc_gen_listing = 1;
			}
			break;

		case 'X':
			/* -Xref : Generate listing through 'cobxref' */
			cobc_gen_listing = 2;
			break;

		case 'D':
			/* -D xx(=yy) : Define variables */
			if (strlen (optarg) > 64) {
				fprintf (stderr, _("Invalid -D parameter\n"));
				exit (1);
			}
			if (!strcasecmp (optarg, "ebug")) {
				fprintf (stderr, _("Warning - passing '%s' to compiler - did you intend to use -debug?\n"), optarg);
				fflush (stderr);
			}
			p = cb_define_list_add (cb_define_list, optarg);
			if (!p) {
				fprintf (stderr, _("Invalid -D parameter\n"));
				exit (1);
			}
			cb_define_list = p;
			break;

#if	0	/* Deprecated */
		case '%':
			/* -MT<file> : Dependency target */
			if (strlen (optarg) > COB_SMALL_MAX) {
				fprintf (stderr, _("Invalid -MT parameter\n"));
				exit (1);
			}
			cb_depend_target = cobc_strdup (optarg);
			break;

		case '@':
			/* -MF<file> : Dependency file */
			if (strlen (optarg) > COB_SMALL_MAX) {
				fprintf (stderr, _("Invalid -MF parameter\n"));
				exit (1);
			}
			cb_depend_file = fopen (optarg, "w");
			if (!cb_depend_file) {
				fflush (stderr);
				perror (optarg);
				fflush (stderr);
			}
			break;
#endif

		case 'I':
			/* -I <xx> : Include/copy directory */
			if (strlen (optarg) > COB_SMALL_MAX) {
				fprintf (stderr, _("Invalid -I parameter\n"));
				exit (1);
			}
			if (stat (optarg, &st) != 0 ||
			    !(S_ISDIR (st.st_mode))) {
				break;
			}
#ifdef	_MSC_VER
			COBC_ADD_STR (cob_include, " /I \"", optarg, "\"");
#else
			COBC_ADD_STR (cob_include, " -I\"", optarg, "\"");
#endif
			cb_include_list = cb_text_list_add (cb_include_list, optarg);
			break;

		case 'L':
			/* -L <xx> : Directory for library search */
			if (strlen (optarg) > COB_SMALL_MAX) {
				fprintf (stderr, _("Invalid -L parameter\n"));
				exit (1);
			}
			if (stat (optarg, &st) != 0 ||
			    !(S_ISDIR (st.st_mode))) {
				break;
			}
#ifdef	_MSC_VER
			COBC_ADD_STR (cob_lib_paths, " /LIBPATH:\"", optarg, "\"");
#else
			COBC_ADD_STR (cob_lib_paths, " -L\"", optarg, "\"");
#endif
			break;

		case 'l':
			/* -l <xx> : Add library to link phase */
			if (strlen (optarg) > COB_SMALL_MAX) {
				fprintf (stderr, _("Invalid -l parameter\n"));
				exit (1);
			}
#ifdef	_MSC_VER
			COBC_ADD_STR (cob_libs, " \"", optarg, "\"");
#else
			COBC_ADD_STR (cob_libs, " -l\"", optarg, "\"");
#endif
			break;

		case 'e':
			/* -e <xx> : Add an extension suffix */
			if (strlen (optarg) > 15) {
				fprintf (stderr, _("Invalid --ext parameter\n"));
				exit (1);
			}
			snprintf (ext, COB_MINI_MAX, ".%s", optarg);
			cb_extension_list = cb_text_list_add (cb_extension_list, ext);
			break;

		case 'K':
			/* -K <xx> : Define literal CALL to xx as static */
			if (strlen (optarg) > 32) {
				fprintf (stderr, _("Invalid -K parameter\n"));
				exit (1);
			}
			cb_static_call_list =
				cb_text_list_add (cb_static_call_list, optarg);
			break;

		case 'k':
			/* -k <xx> : Check for exit after CALL to xx  */
			/* This is to cater for legacy German DIN standard */
			/* Check after CALL if an exit program required */
			/* Not in --help as subject to change and highly specific */
			if (strlen (optarg) > 32) {
				fprintf (stderr, _("Invalid -k parameter\n"));
				exit (1);
			}
			cb_early_exit_list =
				cb_text_list_add (cb_early_exit_list, optarg);
			break;

		case 'M':
			/* -M <xx> : Specify stack (perform) size */
			n = 0;
			n = atoi (optarg);
			if (n < 64 || n >512) {
				fprintf (stderr, _("Invalid --stack-size parameter\n"));
				exit (1);
			}
			cb_stack_size = n;
			break;

		case 'A':
			/* -A <xx> : Add options to C compile phase */
			COBC_ADD_STR (cob_cflags, " ", optarg, NULL);
			break;

		case 'Q':
			/* -Q <xx> : Add options to C link phase */
			COBC_ADD_STR (cob_ldflags, " ", optarg, NULL);
			break;

		case 'w':
			/* -w(xx) : Turn off warnings */
#undef	CB_WARNDEF
#define	CB_WARNDEF(var,name,wall,doc)	var = 0;
#include "warning.def"
#undef	CB_WARNDEF
			break;

		case 'W':
			/* -W : Turn on warnings */
			warningopt = 1;
#undef	CB_WARNDEF
#define	CB_WARNDEF(var,name,wall,doc)	if (wall) var = 1;
#include "warning.def"
#undef	CB_WARNDEF
			break;

		case 'Z':
			/* -W : Turn on all warnings */
			warningopt = 1;
#undef	CB_WARNDEF
#define	CB_WARNDEF(var,name,wall,doc)	var = 1;
#include "warning.def"
#undef	CB_WARNDEF
			break;

		default:
			fprintf (stderr, _("Error: Invalid option detected\n"));
			exit (1);
		}
	}

	/* Exit if list options specified */
	if (exit_option) {
		exit (0);
	}

	/* Load default configuration file if necessary */
	if (cb_config_name == NULL) {
		if (cb_load_std ("default.conf") != 0) {
			fprintf (stderr, _("Error: failed to load the initial config file\n"));
			exit (1);
		}
	}

	/* Check fold combination */
	if (cb_flag_fold_copy_lower && cb_flag_fold_copy_upper) {
		fprintf (stderr, _("Error: Invalid fold-copy combination\n"));
		exit (1);
	}

	/* Check valid tab width */
	if (cb_tab_width < 1 || cb_tab_width > 8) {
		fprintf (stderr, _("Error: Invalid tab-width value - %d\n"),
			 cb_tab_width);
		fflush (stderr);
		exit (1);
	}

	/* Check valid text columns value */
	if (cb_text_column < 72 || cb_text_column > 255) {
		fprintf (stderr, _("Error: Invalid text-column value - %d\n"),
			 cb_text_column);
		fflush (stderr);
		exit (1);
	}

	/* Set relaxed syntax parameters */
	if (cb_flag_relaxed_syntax) {
		cb_relaxed_syntax_check = 1;
		cb_larger_redefines_ok = 1;
		cb_relax_level_hierarchy = 1;
		cb_top_level_occurs_clause = 1;
	}

	/* If C debug, do not strip output */
	if (gflag_set) {
		strip_output = 0;
	}

	if (cb_flag_traceall) {
		cb_flag_trace = 1;
		cb_flag_source_location = 1;
	}

	/* Default extension list */
	cb_extension_list = cb_text_list_add (cb_extension_list, ".CPY");
	cb_extension_list = cb_text_list_add (cb_extension_list, ".CBL");
	cb_extension_list = cb_text_list_add (cb_extension_list, ".COB");
	cb_extension_list = cb_text_list_add (cb_extension_list, ".cpy");
	cb_extension_list = cb_text_list_add (cb_extension_list, ".cbl");
	cb_extension_list = cb_text_list_add (cb_extension_list, ".cob");
	cb_extension_list = cb_text_list_add (cb_extension_list, "");

	return optind;
}

static void
process_env_copy_path (void)
{
	char		*value;
	char		*token;
	char		*p;

	p = getenv ("COBCPY");
	if (p == NULL || strlen (p) == 0) {
		return;
	}

	/* Clone value for destructive strtok */
	value = cobc_strdup (p);

	/* Tokenize for path sep. */
	token = strtok (value, PATHSEPS);
	while (token) {
		cb_include_list = cb_text_list_add (cb_include_list, token);
		token = strtok (NULL, PATHSEPS);
	}

	free (value);
	return;
}

#if	defined(_MSC_VER) || defined(__OS400__)
static void
file_stripext (char *buff)
{
	char	*endp;

	endp = buff + strlen (buff) - 1;
	while (endp > buff) {
		if (*endp == '/' || *endp == '\\') {
			break;
		}
		if (*endp == '.') {
			*endp = 0;
		}
		--endp;
	}
}
#endif

static char *
file_basename (const char *filename)
{
	const char	*startp;
	const char	*endp;
	size_t		len;

	/* Remove directory name */
	startp = strrchr (filename, '/');
	if (startp) {
		startp++;
	} else {
		startp = filename;
	}
	endp = strrchr (startp, '\\');
	if (endp && endp >= startp) {
		startp = ++endp;
	}

	/* Remove extension */
	endp = strrchr (filename, '.');
	if (endp > startp) {
		len = endp - startp;
	} else {
		len = strlen (startp);
	}

	if (len >= basename_len) {
		free (basename_buffer);
		basename_len = len + 16;
		basename_buffer = cobc_malloc (basename_len);
	}
	/* Copy base name */
	memcpy (basename_buffer, startp, len);
	basename_buffer[len] = 0;
	return basename_buffer;
}

static const char *
file_extension (const char *filename)
{
	const char *p;

	p = strrchr (filename, '.');
	if (p) {
		return p + 1;
	} else {
		return "";
	}
}

static char *
cobc_temp_name (const char *ext)
{
#ifdef	_WIN32
	char	buff[COB_MEDIUM_BUFF];
	char	temp[MAX_PATH];

	GetTempPath (MAX_PATH, temp);
	GetTempFileName (temp, "cob", 0, buff);
	DeleteFile (buff);
	/* Replace ".tmp" by EXT */
	strcpy (buff + strlen (buff) - 4, ext);
	return cobc_strdup (buff);
#else
	char	*buff;
	size_t	size;

	size = strlen (cob_tmpdir) + strlen (ext) + 16;
	buff = cobc_malloc (size);
	sprintf (buff, "%s/cob%d_%d%s", cob_tmpdir, cob_process_id,
		 cob_iteration, ext);
	return buff;
#endif
}

static struct filename *
process_filename (const char *filename)
{
	const char	*extension;
	struct filename	*fn;
	struct filename	*ffn;
	char		*basename;
	char		*listptr;
	size_t		fsize;
	struct stat	st;

	fsize = strlen (filename);
	if (fsize > COB_NORMAL_MAX) {
		fprintf (stderr, _("Invalid file name parameter\n"));
		return NULL;
	}
#ifdef	__OS400__
	if (strchr (filename, '.') != NULL)
#endif
	if (stat (filename, &st) != 0) {
		cobc_terminate (filename);
	}
	basename = file_basename (filename);
	if (cobc_check_valid_name (basename, 0)) {
		fprintf (stderr, _("Invalid file base name - %s\n"), basename);
		return NULL;
	}
	fn = cobc_malloc (sizeof (struct filename));
	fn->need_preprocess = 1;
	fn->need_translate = 1;
	fn->need_assemble = 1;
	fn->next = NULL;

	if (!file_list) {
		file_list = fn;
	} else {
		for (ffn = file_list; ffn->next; ffn = ffn->next)
			;
		ffn->next = fn;
	}

	fn->demangle_source =  cb_encode_program_id (basename);
	extension = file_extension (filename);

	/* Check input file type */
	if (strcmp (extension, "i") == 0) {
		/* Already preprocessed */
		fn->need_preprocess = 0;
	} else if (strcmp (extension, "c") == 0 ||
		   strcmp (extension, "s") == 0) {
		/* Already compiled */
		fn->need_preprocess = 0;
		fn->need_translate = 0;
#ifdef	_MSC_VER
	} else if (strcmp (extension, "obj") == 0 ||
		   strcmp (extension, "lib") == 0) {
#elif	defined(__OS400__)
	} else if (extension[0] == 0) {
#else
	} else if (strcmp (extension, "o") == 0 ||
		   strcmp (extension, "a") == 0 ||
		   strcmp (extension, "so") == 0 ||
		   strcmp (extension, "sl") == 0) {
#endif
		/* Already assembled */
		fn->need_preprocess = 0;
		fn->need_translate = 0;
		fn->need_assemble = 0;
	}

	/* Set source filename */
	fn->source = cobc_strdup (filename);

	/* Set preprocess filename */
	if (!fn->need_preprocess) {
		fn->preprocess = cobc_strdup (fn->source);
	} else if (output_name && cb_compile_level == CB_LEVEL_PREPROCESS) {
		fn->preprocess = cobc_strdup (output_name);
	} else if (save_csrc || save_temps ||
		   cb_compile_level == CB_LEVEL_PREPROCESS) {
		fn->preprocess = cobc_stradd_dup (basename, ".i");
	} else {
		fn->preprocess = cobc_temp_name (".cob");
	}

	/* Set translate filename */
	if (!fn->need_translate) {
		fn->translate = cobc_strdup (fn->source);
	} else if (output_name && cb_compile_level == CB_LEVEL_TRANSLATE) {
		fn->translate = cobc_strdup (output_name);
	} else if (save_csrc || save_temps ||
		   cb_compile_level == CB_LEVEL_TRANSLATE) {
		fn->translate = cobc_stradd_dup (basename, ".c");
	} else {
		fn->translate = cobc_temp_name (".c");
	}

	/* Set storage filename */
	if (fn->need_translate) {
		fn->trstorage = cobc_stradd_dup (fn->translate, ".h");
	}

	/* Set object filename */
	if (!fn->need_assemble) {
		fn->object = cobc_strdup (fn->source);
	} else if (output_name && cb_compile_level == CB_LEVEL_ASSEMBLE) {
		fn->object = cobc_strdup (output_name);
	} else if (save_temps || cb_compile_level == CB_LEVEL_ASSEMBLE) {
#if	defined(_MSC_VER) || defined(__OS400__)
		fn->object = cobc_stradd_dup (basename, ".obj");
#else
		fn->object = cobc_stradd_dup (basename, ".o");
#endif
	} else {
#if	defined(_MSC_VER) || defined(__OS400__)
		fn->object = cobc_stradd_dup (basename, ".obj");
#else
		fn->object = cobc_temp_name (".o");
#endif
	}

	/* Set listing filename */
	if (cobc_gen_listing == 1) {
		if (cobc_list_file) {
			fn->listing_file = cobc_list_file;
		} else if (cobc_list_dir) {
			fsize = strlen (cobc_list_dir) + strlen (basename) + 8U;
			listptr = cobc_malloc (fsize);
			snprintf (listptr, fsize, "%s/%s.lst",
				  cobc_list_dir, basename);
			fn->listing_file = listptr;
		} else {
			fn->listing_file = cobc_stradd_dup (basename, ".lst");
		}
	} else if (cobc_gen_listing > 1) {
		fn->listing_file = cobc_stradd_dup (basename, ".xrf");
	}

	cob_iteration++;

	return fn;
}

#ifdef	__OS400__
static int
process (char *cmd)
{
	char	*buffptr;
	char	*name = NULL;
	char	*objname = NULL;
	char	*cobjname = NULL;
	char	*token;
	char	*incl[100];
	char	*defs[100];
	char	*objs[100];
	char	*libs[100];
	char	*optc[100];
	char	*optl[100];
	int	nincl = 0;
	int	ndefs = 0;
	int	nobjs = 0;
	int	nlibs = 0;
	int	noptc = 0;
	int	noptl = 0;
	int	dbg = 0;
	int	comp_only = 0;
	int	shared = 0;
	int	optimize = 0;
	int	i;
	int	len;
	int	ret;

	if (verbose_output) {
		cobc_cmd_print (cmd);
	}
	if (gflag_set) {
		dbg = 1;
	}
	token = strtok (cmd, " ");
	if (token != NULL) {
		/* Skip C compiler */
		token = strtok (NULL, " ");
	}
	for (; token; token = strtok (NULL, " ")) {
		if (*token != '-') {
			len = strlen (token);
			if (*token == '"') {
				len -= 2;
				++token;
				token[len] = 0;
			}
			if (token[len-2] == '.' && token[len-1] == 'c') {
				/* C source */
				name = token;
				continue;
			}
			/* Assuming module */
			objs[nobjs++] = token;
			continue;
		}
		++token;
		switch (*token) {
		case 'c':
			comp_only = 1;
			break;
		case 'I':
			++token;
			if (*token == 0) {
				token = strtok (NULL, " ");
			}
			if (*token == '"') {
				++token;
				token[strlen(token) - 1] = 0;
			}
			incl[nincl++] = token;
			break;
		case 'D':
			++token;
			if (*token == 0) {
				token = strtok (NULL, " ");
			}
			if (*token == '"') {
				++token;
				token[strlen(token) - 1] = 0;
			}
			defs[ndefs++] = token;
			break;
		case 'A':
			++token;
			optc[noptc++] = token;
			break;
		case 'Q':
			++token;
			optl[noptl++] = token;
			break;
		case 'o':
			++token;
			if (*token == 0) {
				token = strtok (NULL, " ");
			}
			if (*token == '"') {
				++token;
				token[strlen(token) - 1] = 0;
			}
			objname = token;
			break;
		case 'l':
			++token;
			if (*token == 0) {
				token = strtok (NULL, " ");
			}
			libs[nlibs++] = token;
			break;
		case 'G':
			shared = 1;
			break;
		case 'g':
			dbg = 1;
			break;
		case 'O':
			optimize = 1;
			break;
		default:
			fprintf (stderr, _("Unknown option ignored:\t%s\n"),
				 token - 1);
			fflush (stderr);
		}
	}

	buffptr = cobc_malloc (COB_LARGE_BUFF);
	if (name != NULL) {
		/* Requires compilation */
		if (objname == NULL) {
			cobjname = file_basename (name);
		} else {
			cobjname = objname;
		}
		sprintf(buffptr, "CRTCMOD MODULE(%s) SRCSTMF('%s') ",
			cobjname, name);
		if (nincl > 0) {
			strcat (buffptr, "INCDIR(");
			for (i = 0; i < nincl; ++i) {
				if (i != 0) {
					strcat (buffptr, " ");
				}
				strcat (buffptr, "'");
				strcat (buffptr, incl[i]);
				strcat (buffptr, "' ");
			}
			strcat (buffptr, ") ");
		}
		if (ndefs > 0) {
			strcat (buffptr, "DEFINE(");
			for (i = 0; i < ndefs; ++i) {
				if (i != 0) {
					strcat (buffptr, " ");
				}
				strcat (buffptr, "'");
				strcat (buffptr, defs[i]);
				strcat (buffptr, "' ");
			}
			strcat (buffptr, ") ");
		}
		strcat (buffptr, "SYSIFCOPT(*IFSIO)");
		for (i = 0; i < noptc; ++i) {
			strcat (buffptr, " ");
			strcat (buffptr, optc[i]);
		}
		if (optimize) {
			strcat (buffptr, " OPTIMIZE(40)");
		}
		if (dbg) {
			strcat (buffptr, " DBGVIEW(*ALL)");
		}
		if (cobc_gen_listing) {
			strcat (buffptr, " OUTPUT(*PRINT)");
		}
		if (verbose_output) {
			cobc_cmd_print (buffptr);
		}
		ret = system (buffptr);
		if (verbose_output) {
			fprintf (stderr, _("Return status:\t%d\n"), ret);
			fflush (stderr);
		}
		if (comp_only || ret != 0) {
			free (buffptr);
			return ret;
		}
	}
	if (objname == NULL) {
		if (name != NULL) {
			objname = cobjname;
		} else if (nobjs > 0) {
			objname = objs[0];
		} else {
			objname = (char *)"AOUT";
		}
	}
	if (shared) {
		sprintf(buffptr, "CRTSRVPGM SRVPGM(%s) MODULE(", objname);
	} else {
		sprintf(buffptr, "CRTPGM PGM(%s) MODULE(", objname);
	}
	if (name != NULL) {
		strcat (buffptr, cobjname);
	}
	for (i = 0; i < nobjs; ++i) {
		if (i != 0 || name != NULL) {
			strcat (buffptr, " ");
		}
		strcat (buffptr, objs[i]);
	}
	strcat (buffptr, ")");
	if (nlibs > 0) {
		strcat (buffptr, " BNDSRVPGM(");
		for (i = 0; i < nlibs; ++i) {
			if (i != 0) {
				strcat (buffptr, " ");
			}
			strcat (buffptr, libs[i]);
		}
		strcat (buffptr, ")");
	}
	for (i = 0; i < noptl; ++i) {
		strcat (buffptr, " ");
		strcat (buffptr, optl[i]);
	}
	if (shared) {
		strcat (buffptr, " EXPORT(*ALL)");
	}
	if (verbose_output) {
		cobc_cmd_print (buffptr);
	}
	ret = system (buffptr);
	if (verbose_output) {
		fprintf (stderr, _("Return status:\t%d\n"), ret);
		fflush (stderr);
	}
	free (buffptr);
	return ret;
}

#else

static int
process (const char *cmd)
{
	char	*p;
	char	*buffptr;
	size_t	clen;
	int	ret;

	if (strchr (cmd, '$') == NULL) {
		if (verbose_output) {
			cobc_cmd_print (cmd);
		}
		ret = system (cmd);
		if (verbose_output) {
			fprintf (stderr, _("Return status:\t%d\n"), ret);
			fflush (stderr);
		}
		return ret;
	}
	clen = strlen (cmd) + 64;
	buffptr = cobc_malloc (clen);
	p = buffptr;
	/* Quote '$' */
	for (; *cmd; ++cmd) {
		if (*cmd == '$') {
			p += sprintf (p, "\\$");
		} else {
			*p++ = *cmd;
		}
	}
	*p = 0;

	if (verbose_output) {
		cobc_cmd_print (buffptr);
	}
	ret = system (buffptr);
	if (verbose_output) {
		fprintf (stderr, _("Return status:\t%d\n"), ret);
		fflush (stderr);
	}
	free (buffptr);
	return ret;
}
#endif

/* Preprocess source */

static int
preprocess (struct filename *fn)
{
#if	0	/* Deprecated */
	struct cb_text_list	*l;
#endif
	int			save_source_format;
	int			save_fold_upper;
	int			save_fold_lower;
	int			ret;

	errorcount = 0;

	ppout = fopen (fn->preprocess, "w");
	if (!ppout) {
		cobc_terminate (fn->preprocess);
	}

	if (ppopen (fn->source, NULL) != 0) {
		fclose (ppout);
		ppout = NULL;
		if (fn->preprocess) {
			unlink (fn->preprocess);
		}
		exit (1);
	}

	if (verbose_output) {
		fprintf (stderr, _("Preprocessing:\t%s to %s\n"),
			 fn->source, fn->preprocess);
		fflush (stderr);
	}

	if (cobc_gen_listing && !cobc_list_file) {
		cb_listing_file = fopen (fn->listing_file, "w");
		if (!cb_listing_file) {
			cobc_terminate (fn->listing_file);
		}
	}

	/* Reset plex variables */
	plex_clear_vars ();

	/* Save default flags in case program directives change them */
	save_source_format = cb_source_format;
	save_fold_upper = cb_flag_fold_copy_upper;
	save_fold_lower = cb_flag_fold_copy_lower;

	/* Preprocess */
	ppparse ();

	/* Restore default flags */
	cb_source_format = save_source_format;
	cb_flag_fold_copy_upper = save_fold_upper;
	cb_flag_fold_copy_lower = save_fold_lower;

	/* Check dangling directive */
	plex_check_dangling ();

	if (ppin) {
		fclose (ppin);
		ppin = NULL;
	}

	if (ppout) {
		fclose (ppout);
		ppout = NULL;
	}

	if (cobc_gen_listing && !cobc_list_file) {
		fclose (cb_listing_file);
		if (cobc_gen_listing > 1) {
			sprintf (cobc_buffer, "cobxref %s -R", fn->listing_file);
			if (verbose_output) {
				cobc_cmd_print (cobc_buffer);
			}
			ret = system (cobc_buffer);
			if (verbose_output) {
				fprintf (stderr, _("Return status:\t%d\n"), ret);
				fflush (stderr);
			}
			if (ret) {
				fprintf (stderr, _("'cobxref' execution unsuccessful\n"));
#ifdef	_WIN32
				fprintf (stderr, _("Check that 'cobxref' is in %PATH%\n"));
#else
				fprintf (stderr, _("Check that 'cobxref' is in $PATH\n"));
#endif
				fprintf (stderr, _("No listing produced\n"));
				fflush (stderr);
			}
			unlink (fn->listing_file);
		}
		cb_listing_file = NULL;
	}

	if (verbose_output) {
		fprintf (stderr, _("Return status:\t%d\n"), errorcount);
		fflush (stderr);
	}

	if (errorcount > 0) {
		return 1;
	}

#if	0	/* Deprecated */
	/* Output dependency list */
	if (cb_depend_file) {
		if (!cb_depend_target) {
			fputs (_("-MT must be given to specify target file\n"),
			       stderr);
			exit (1);
		}
		fprintf (cb_depend_file, "%s: \\\n", cb_depend_target);
		for (l = cb_depend_list; l; l = l->next) {
			fprintf (cb_depend_file, " %s%s\n",
				 l->text, l->next ? " \\" : "");
		}
		for (l = cb_depend_list; l; l = l->next) {
			fprintf (cb_depend_file, "%s:\n", l->text);
		}
		fclose (cb_depend_file);
	}
#endif

	return 0;
}

/* Reverse the list of programs */

static struct cb_program *
program_list_reverse (struct cb_program *p)
{
	struct cb_program	*next;
	struct cb_program	*last;

	last = NULL;
	for (; p; p = next) {
		next = p->next_program;
		p->next_program = last;
		last = p;
	}
	return last;
}

/* Create single-element C source */

static int
process_translate (struct filename *fn)
{
	struct cb_program	*p;
	struct cb_program	*q;
	struct cb_program	*r;
	struct nested_list	*nlp;
	struct handler_struct	*hstr1;
	struct handler_struct	*hstr2;
	struct local_filename	*lf;
	int			ret;
	int			i;

	/* Initialize */
	cb_source_file = NULL;
	cb_source_line = 0;
	cb_init_constants ();
	cb_init_reserved ();

	/* Open the input file */
	yyin = fopen (fn->preprocess, "r");
	if (!yyin) {
		cobc_terminate (fn->preprocess);
	}

	/* Parse */
	if (verbose_output) {
		fprintf (stderr, _("Parsing:\t%s\n"), fn->preprocess);
		fflush (stderr);
	}
	ret = yyparse ();
	fclose (yyin);
	yyin = NULL;

	if (verbose_output) {
		fprintf (stderr, _("Return status:\t%d\n"), ret);
		fflush (stderr);
	}

	if (ret) {
		return ret;
	}
	if (cb_flag_syntax_only || current_program->entry_list == NULL) {
		return 0;
	}

	/* Set up USE GLOBAL handlers */
	p = current_program;
	for (q = p; q; q = q->next_program) {
		q->global_file_list = cb_list_reverse (q->global_file_list);
		if (q->nested_level) {
			for (r = q->next_program; r; r = r->next_program) {
				if (r->nested_level >= q->nested_level) {
					continue;
				}
				for (i = COB_OPEN_INPUT; i <= COB_OPEN_EXTEND; ++i) {
					hstr1 = &q->global_handler[i];
					hstr2 = &r->global_handler[i];
					if (!hstr1->handler_label &&
					    hstr2->handler_label &&
					    hstr2->handler_label->flag_global) {
						hstr1->handler_label = hstr2->handler_label;
						hstr1->handler_prog = r;
					}
				}
				if (!r->nested_level) {
					break;
				}
			}
		}
	}

	if (verbose_output) {
		fprintf (stderr, _("Translating:\t%s to %s\n"),
			 fn->preprocess, fn->translate);
		fflush (stderr);
	}

	/* Open the output file */
	yyout = fopen (fn->translate, "w");
	if (!yyout) {
		cobc_terminate (fn->translate);
	}

	/* Open the common storage file */
	cb_storage_file_name = fn->trstorage;
	cb_storage_file = fopen (cb_storage_file_name, "w");
	if (!cb_storage_file) {
		cobc_terminate (cb_storage_file_name);
	}

	p = program_list_reverse (current_program);

	/* Set up local storage files */
	ret = 1;
	for (q = p; q; q = q->next_program, ret++) {
		lf = cobc_malloc (sizeof(struct local_filename));
		lf->local_name = cobc_malloc (strlen (fn->translate) + 9);
		if (q == p && !q->next_program) {
			sprintf (lf->local_name, "%s.l.h", fn->translate);
		} else {
			sprintf (lf->local_name, "%s.l%d.h", fn->translate, ret);
		}
		lf->local_fp = fopen (lf->local_name, "w");
		if (!lf->local_fp) {
			cobc_terminate (lf->local_name);
		}
		q->local_storage_file = lf->local_fp;
		q->local_storage_name = lf->local_name;
		lf->next = fn->localfile;
		fn->localfile = lf;
	}

	/* Entries for COMMON programs */
	for (q = p; q; q = q->next_program) {
		i = q->nested_level;
		for (nlp = q->common_prog_list; nlp; nlp = nlp->next) {
			for (r = q->next_program; r; r = r->next_program) {
				if (r->nested_level <= i) {
					break;
				}
				cb_insert_common_prog (r, nlp->nested_prog);
			}
		}
	}

	/* Translate to C */
	codegen (p, 0);

	/* Close files */
	fclose (cb_storage_file);
	fclose (yyout);
	yyout = NULL;
	for (q = p; q; q = q->next_program) {
		fclose (q->local_storage_file);
	}
	return 0;
}

/* Create single-element assembly source */

static int
process_compile (struct filename *fn)
{
	char	*name;
	size_t	bufflen;

	if (output_name) {
		name = output_name;
	} else {
		name = file_basename (fn->source);
#ifndef	_MSC_VER
		strcat (name, ".s");
#endif
	}

	bufflen = strlen (cob_cc) + strlen (cob_cflags)
			+ strlen (name) + strlen (fn->translate)
			+ strlen (cob_include) + 64;

	if (bufflen >= cobc_buffer_size) {
		free (cobc_buffer);
		cobc_buffer_size = bufflen + 16;
		cobc_buffer = cobc_malloc (cobc_buffer_size);
	}


#ifdef	_MSC_VER
	sprintf (cobc_buffer, gflag_set ?
		"%s /c %s %s /Od /MDd /Zi /FR /c /Fa\"%s\" /Fo\"%s\" %s" :
		"%s /c %s %s /MD /c /Fa\"%s\" /Fo\"%s\" %s",
			cob_cc, cob_cflags, cob_include, name,
			name, fn->translate);
#else
	sprintf (cobc_buffer, "%s -S -o %s %s %s %s", cob_cc, name,
			cob_cflags, cob_include, fn->translate);
#endif
	return process (cobc_buffer);
}

/* Create single-element assembled object */

static int
process_assemble (struct filename *fn)
{
	size_t		bufflen;

	bufflen = strlen (cob_cc) + strlen (cob_cflags)
			+ strlen (fn->object) + strlen (fn->translate)
			+ strlen (cob_include) + strlen (COB_PIC_FLAGS) + 64;

	if (bufflen >= cobc_buffer_size) {
		free (cobc_buffer);
		cobc_buffer_size = bufflen + 16;
		cobc_buffer = cobc_malloc (cobc_buffer_size);
	}

#ifdef	_MSC_VER
	sprintf (cobc_buffer, gflag_set ?
		"%s /c %s %s /Od /MDd /Zi /FR /Fo\"%s\" %s" :
		"%s /c %s %s /MD /Fo\"%s\" \"%s\"",
			cob_cc, cob_cflags, cob_include,
			fn->object, fn->translate);
#elif defined(__OS400__)
	char	*name;

	name = (char *) fn->translate;
	if (name[0] != '/') {
		char	*p;

		p = cobc_malloc (COB_LARGE_BUFF);
		getcwd (p, COB_LARGE_BUFF);
		strcat (p, "/");
		strcat (p, name);
		name = p;
	}
	file_stripext ((char *) fn->object);
	sprintf (cobc_buffer, "%s -c %s %s -o %s %s",
		 cob_cc, cob_cflags, cob_include,
		 fn->object, name);
#else
	if (cb_compile_level == CB_LEVEL_MODULE ||
	    cb_compile_level == CB_LEVEL_LIBRARY) {
		sprintf (cobc_buffer, "%s -c %s %s %s -o \"%s\" \"%s\"",
			 cob_cc, cob_cflags, cob_include,
			 COB_PIC_FLAGS, fn->object, fn->translate);
	} else {
		sprintf (cobc_buffer, "%s -c %s %s -o \"%s\" \"%s\"",
			 cob_cc, cob_cflags, cob_include,
			 fn->object, fn->translate);
	}
#endif
	return process (cobc_buffer);
}

/* Create single-element loadable object without intermediate stages */

static int
process_module_direct (struct filename *fn)
{
	char	*name;
	size_t	bufflen;
	int	ret;

	if (output_name) {
		name = output_name_buff;
		strcpy (name, output_name);
#if	defined(_MSC_VER) || defined(__OS400__)
		file_stripext (name);
#else
		if (strchr (output_name, '.') == NULL) {
			strcat (name, ".");
			strcat (name, COB_MODULE_EXT);
		}
#endif
	} else {
		name = file_basename (fn->source);
#if	!defined(_MSC_VER) && !defined(__OS400__)
		strcat (name, ".");
		strcat (name, COB_MODULE_EXT);
#endif
	}

	bufflen = strlen (cob_cc) + strlen (cob_cflags)
			+ strlen (COB_EXPORT_DYN) + strlen (COB_SHARED_OPT)
			+ strlen (name) + strlen (fn->translate)
			+ strlen (cob_libs) + strlen (cob_ldflags)
			+ strlen (cob_include)
#ifdef	_MSC_VER
			+ strlen (manilink)
#endif
			+ strlen (cob_lib_paths) + strlen (COB_PIC_FLAGS) + 128;

	if (bufflen >= cobc_buffer_size) {
		free (cobc_buffer);
		cobc_buffer_size = bufflen + 16;
		cobc_buffer = cobc_malloc (cobc_buffer_size);
	}

#ifdef	_MSC_VER
	sprintf (cobc_buffer, gflag_set ?
		"%s %s %s /Od /MDd /LDd /Zi /FR /Fe\"%s\" /Fo\"%s\" %s \"%s\" %s %s %s" :
		"%s %s %s /MD /LD /Fe\"%s\" /Fo\"%s\" %s \"%s\" %s %s %s",
			cob_cc, cob_cflags, cob_include, name, name,
			cob_ldflags, fn->translate, cob_libs,
			manilink, cob_lib_paths);
	ret = process (cobc_buffer);
#if	_MSC_VER >= 1400
	/* Embedding manifest */
	if (ret == 0) {
		sprintf (cobc_buffer,
			 "%s /manifest \"%s.dll.manifest\" /outputresource:\"%s.dll\";#2",
			 manicmd, name, name);
		ret = process (cobc_buffer);
		sprintf (cobc_buffer, "%s.dll.manifest", name);
		cobc_check_action (cobc_buffer);
	}
#endif
	sprintf (cobc_buffer, "%s.exp", name);
	cobc_check_action (cobc_buffer);
	sprintf (cobc_buffer, "%s.lib", name);
	cobc_check_action (cobc_buffer);
#else	/* _MSC_VER */
#ifdef	__OS400__
	if (fn->translate[0] != '/') {
		char	*p;

		p = cobc_malloc (COB_LARGE_BUFF);
		getcwd (p, COB_LARGE_BUFF);
		strcat (p, "/");
		strcat (p, fn->translate);
		fn->translate = p;
	}
#endif
	sprintf (cobc_buffer, "%s %s %s %s %s %s %s -o \"%s\" \"%s\" %s %s",
		 cob_cc, cob_cflags, cob_include, COB_SHARED_OPT,
		 cob_ldflags, COB_PIC_FLAGS, COB_EXPORT_DYN, name,
		 fn->translate, cob_libs, cob_lib_paths);
	ret = process (cobc_buffer);
#ifdef	COB_STRIP_CMD
	if (strip_output && ret == 0) {
		sprintf (cobc_buffer, "%s \"%s\"", COB_STRIP_CMD, name);
		ret = process (cobc_buffer);
	}
#endif
#endif	/* _MSC_VER */
	return ret;
}

/* Create single-element loadable object */

static int
process_module (struct filename *fn)
{
	char	*name;
	size_t	bufflen;
	int	ret;

	if (output_name) {
		name = output_name_buff;
		strcpy (name, output_name);
#if	defined(_MSC_VER) || defined(__OS400__)
		file_stripext (name);
#else
		if (strchr (output_name, '.') == NULL) {
			strcat (name, ".");
			strcat (name, COB_MODULE_EXT);
		}
#endif
	} else {
		name = file_basename (fn->source);
#if	!defined(_MSC_VER) && !defined(__OS400__)
		strcat (name, ".");
		strcat (name, COB_MODULE_EXT);
#endif
	}

	bufflen = strlen (cob_cc) + strlen (cob_ldflags)
			+ strlen (COB_EXPORT_DYN) + strlen (COB_SHARED_OPT)
			+ strlen (name) + strlen (fn->object) + strlen (cob_libs)
#ifdef	_MSC_VER
			+ strlen (manilink)
#endif
			+ strlen (cob_lib_paths) + strlen (COB_PIC_FLAGS) + 128;

	if (bufflen >= cobc_buffer_size) {
		free (cobc_buffer);
		cobc_buffer_size = bufflen + 16;
		cobc_buffer = cobc_malloc (cobc_buffer_size);
	}

#ifdef	_MSC_VER
	sprintf (cobc_buffer, gflag_set ?
		"%s /Od /MDd /LDd /Zi /FR /Fe\"%s\" %s \"%s\" %s %s %s" :
		"%s /MD /LD /Fe\"%s\" %s \"%s\" %s %s %s",
		cob_cc, name, cob_ldflags, fn->object, cob_libs,
		manilink, cob_lib_paths);
	ret = process (cobc_buffer);
#if	_MSC_VER >= 1400
	/* Embedding manifest */
	if (ret == 0) {
		sprintf (cobc_buffer,
			 "%s /manifest \"%s.dll.manifest\" /outputresource:\"%s.dll\";#2",
			 manicmd, name, name);
		ret = process (cobc_buffer);
		sprintf (cobc_buffer, "%s.dll.manifest", name);
		cobc_check_action (cobc_buffer);
	}
#endif
	sprintf (cobc_buffer, "%s.exp", name);
	cobc_check_action (cobc_buffer);
	sprintf (cobc_buffer, "%s.lib", name);
	cobc_check_action (cobc_buffer);
#else	/* _MSC_VER */
	sprintf (cobc_buffer, "%s %s %s %s %s -o \"%s\" \"%s\" %s %s",
		 cob_cc, COB_SHARED_OPT, cob_ldflags, COB_PIC_FLAGS,
		 COB_EXPORT_DYN, name, fn->object, cob_libs, cob_lib_paths);
	ret = process (cobc_buffer);
#ifdef	COB_STRIP_CMD
	if (strip_output && ret == 0) {
		sprintf (cobc_buffer, "%s %s", COB_STRIP_CMD, name);
		ret = process (cobc_buffer);
	}
#endif
#endif	/* _MSC_VER */
	return ret;
}

/* Create multi-element loadable object */

static int
process_library (struct filename *l)
{
	char		*objsptr;
	struct filename	*f;
	char		*name;
	size_t		bufflen;
	int		ret;

	bufflen = 0;
	for (f = l; f; f = f->next) {
		bufflen += strlen (f->object) + 8;
	}
	objsptr = cobc_malloc (bufflen);
	for (f = l; f; f = f->next) {
		strcat (objsptr, "\"");
		strcat (objsptr, f->object);
		strcat (objsptr, "\" ");
	}

	if (output_name) {
		name = output_name_buff;
		strcpy (name, output_name);
#if	defined(_MSC_VER) || defined(__OS400__)
		file_stripext (name);
#else
		if (strchr (output_name, '.') == NULL) {
			strcat (name, ".");
			strcat (name, COB_MODULE_EXT);
		}
#endif
	} else {
		name = file_basename (l->source);
#if	!defined(_MSC_VER) && !defined(__OS400__)
		strcat (name, ".");
		strcat (name, COB_MODULE_EXT);
#endif
	}

	bufflen = strlen (cob_cc) + strlen (cob_ldflags)
			+ strlen (COB_EXPORT_DYN) + strlen (COB_SHARED_OPT)
			+ strlen (name) + strlen (objsptr) + strlen (cob_libs)
#ifdef	_MSC_VER
			+ strlen (manilink)
#endif
			+ strlen (cob_lib_paths) + strlen (COB_PIC_FLAGS) + 64;
	if (bufflen >= cobc_buffer_size) {
		free (cobc_buffer);
		cobc_buffer_size = bufflen + 16;
		cobc_buffer = cobc_malloc (cobc_buffer_size);
	}

#ifdef	_MSC_VER
	sprintf (cobc_buffer, gflag_set ?
		"%s /Od /MDd /LDd /Zi /FR /Fe\"%s\" %s %s %s %s %s" :
		"%s /MD /LD /Fe\"%s\" %s %s %s %s %s",
		cob_cc, name, cob_ldflags, objsptr, cob_libs,
		manilink, cob_lib_paths);
	ret = process (cobc_buffer);
#if	_MSC_VER >= 1400
	/* Embedding manifest */
	if (ret == 0) {
		sprintf (cobc_buffer,
			 "%s /manifest \"%s.dll.manifest\" /outputresource:\"%s.dll\";#2",
			 manicmd, name, name);
		ret = process (cobc_buffer);
		sprintf (cobc_buffer, "%s.dll.manifest", name);
		cobc_check_action (cobc_buffer);
	}
	sprintf (cobc_buffer, "%s.exp", name);
	cobc_check_action (cobc_buffer);
	sprintf (cobc_buffer, "%s.lib", name);
	cobc_check_action (cobc_buffer);
#endif
#else	/* _MSC_VER */
	sprintf (cobc_buffer, "%s %s %s %s %s -o \"%s\" %s %s %s",
		 cob_cc, COB_SHARED_OPT, cob_ldflags, COB_PIC_FLAGS,
		 COB_EXPORT_DYN, name, objsptr, cob_libs, cob_lib_paths);
	ret = process (cobc_buffer);
#ifdef	COB_STRIP_CMD
	if (strip_output && ret == 0) {
		sprintf (cobc_buffer, "%s \"%s\"", COB_STRIP_CMD, name);
		ret = process (cobc_buffer);
	}
#endif
#endif	/* _MSC_VER */
	return ret;
}

/* Create executable */

static int
process_link (struct filename *l)
{
	char		*objsptr;
	struct filename	*f;
	char		*name;
	size_t		bufflen;
	int		ret;

	bufflen = 0;
	for (f = l; f; f = f->next) {
		bufflen += strlen (f->object) + 8;
	}
	objsptr = cobc_malloc (bufflen);
	for (f = l; f; f = f->next) {
#ifdef	__OS400__
		file_stripext ((char *) f->object);
		strcat (objsptr, f->object);
		strcat (objsptr, " ");
#else
		strcat (objsptr, "\"");
		strcat (objsptr, f->object);
		strcat (objsptr, "\" ");
#endif
	}

	if (output_name) {
#if	defined(_MSC_VER) || defined(__OS400__)
		name = cobc_strdup (output_name);
		file_stripext (name);
#else
		name = output_name;
#endif
	} else {
		name = file_basename (l->source);
	}

	bufflen = strlen (cob_cc) + strlen (cob_ldflags)
			+ strlen (COB_EXPORT_DYN) + strlen (name)
			+ strlen (objsptr) + strlen (cob_libs)
#ifdef	_MSC_VER
			+ strlen (manilink)
#endif
			+ strlen (cob_lib_paths) + 64;
	if (bufflen >= cobc_buffer_size) {
		free (cobc_buffer);
		cobc_buffer_size = bufflen + 16;
		cobc_buffer = cobc_malloc (cobc_buffer_size);
	}

#ifdef	_MSC_VER
	sprintf (cobc_buffer, gflag_set ?
		"%s /Od /MDd /Zi /FR /Fe\"%s\" %s %s %s %s %s" :
		"%s /MD /Fe\"%s\" %s %s %s %s %s",
		cob_cc, name, cob_ldflags, objsptr, cob_libs,
		manilink, cob_lib_paths);
	ret = process (cobc_buffer);
#if	_MSC_VER >= 1400
	/* Embedding manifest */
	if (ret == 0) {
		sprintf (cobc_buffer,
			 "%s /manifest \"%s.exe.manifest\" /outputresource:\"%s.exe\";#1",
			 manicmd, name, name);
		ret = process (cobc_buffer);
		sprintf (cobc_buffer, "%s.exe.manifest", name);
		cobc_check_action (cobc_buffer);
	}
#endif
#else	/* _MSC_VER */
	sprintf (cobc_buffer, "%s %s %s -o \"%s\" %s %s %s",
		 cob_cc, cob_ldflags, COB_EXPORT_DYN, name,
		 objsptr, cob_libs, cob_lib_paths);

	ret = process (cobc_buffer);
#ifdef	__hpux
	if (ret == 0) {
		sprintf (cobc_buffer, "chatr -s +s enable \"%s%s\" 1>/dev/null 2>&1",
			 name, COB_EXEEXT);
		process (cobc_buffer);
	}
#endif
#ifdef	COB_STRIP_CMD
	if (strip_output && ret == 0) {
		sprintf (cobc_buffer, "%s \"%s%s\"",
			 COB_STRIP_CMD, name, COB_EXEEXT);
		ret = process (cobc_buffer);
	}
#endif
#endif	/* _MSC_VER */
	free (objsptr);
	return ret;
}

/* Main function */
int
main (int argc, char *argv[])
{
	struct filename		*fn;
	char			*p;
	size_t			size;
	int			status = 1;
	int			year;
	int			day;
	char			month[32];
#if	defined(HAVE_SIGNAL_H) && defined(HAVE_SIGACTION)
	struct sigaction	sa;
#endif

	file_list = NULL;
	cb_listing_file = NULL;
	ppin = NULL;
	ppout = NULL;
	yyin = NULL;
	yyout = NULL;

#ifdef	HAVE_SIGNAL_H

#ifdef	HAVE_SIGACTION
	sa.sa_handler = cobc_sig_handler;
	sa.sa_flags = 0;
#ifdef	SIGINT
	(void)sigaction (SIGINT, NULL, &intsig);
	if (intsig.sa_handler != SIG_IGN) {
		(void)sigemptyset (&sa.sa_mask);
		(void)sigaction (SIGINT, &sa, NULL);
	}
#endif
#ifdef	SIGHUP
	(void)sigaction (SIGHUP, NULL, &hupsig);
	if (hupsig.sa_handler != SIG_IGN) {
		(void)sigemptyset (&sa.sa_mask);
		(void)sigaction (SIGHUP, &sa, NULL);
	}
#endif
#ifdef	SIGQUIT
	(void)sigaction (SIGQUIT, NULL, &qutsig);
	if (qutsig.sa_handler != SIG_IGN) {
		(void)sigemptyset (&sa.sa_mask);
		(void)sigaction (SIGQUIT, &sa, NULL);
	}
#endif
#ifdef	SIGSEGV
	/* Take direct control of segementation violation */
	(void)sigemptyset (&sa.sa_mask);
	(void)sigaction (SIGSEGV, &sa, NULL);
#endif

#else

#ifdef	SIGINT
	if ((intsig = signal (SIGINT, cobc_sig_handler)) == SIG_IGN) {
		(void)signal (SIGINT, SIG_IGN);
	}
#endif
#ifdef	SIGHUP
	if ((hupsig = signal (SIGHUP, cobc_sig_handler)) == SIG_IGN) {
		(void)signal (SIGHUP, SIG_IGN);
	}
#endif
#ifdef	SIGQUIT
	if ((qutsig = signal (SIGQUIT, cobc_sig_handler)) == SIG_IGN) {
		(void)signal (SIGQUIT, SIG_IGN);
	}
#endif
	/* Take direct control of segmentation violation */
#ifdef	SIGSEGV
	(void)signal (SIGSEGV, cobc_sig_handler);
#endif
#endif

#endif

	cb_saveargc = argc;
	cb_saveargv = argv;

	/* General buffers */
	cobc_buffer = cobc_malloc (COB_LARGE_BUFF);
	cobc_buffer_size = COB_LARGE_MAX;
	basename_buffer = cobc_malloc (COB_MINI_BUFF);
	basename_len = COB_MINI_MAX - 16;

	cob_libs = cobc_malloc (COB_SMALL_BUFF);
	cob_lib_paths = cobc_malloc (COB_SMALL_BUFF);
	cob_cflags = cobc_malloc (COB_MINI_BUFF);
	cob_ldflags = cobc_malloc (COB_MINI_BUFF);
	cob_include = cobc_malloc (COB_MINI_BUFF);
	cob_cflags_size = COB_MINI_MAX;
	cob_libs_size = COB_SMALL_MAX;
	cob_lib_paths_size = COB_SMALL_MAX;
	cob_include_size = COB_MINI_MAX;
	cob_ldflags_size = COB_MINI_MAX;

	cb_source_file = NULL;
	save_temps_dir = NULL;
	cobc_temp_buffer = NULL;
	cb_id = 1;
	cb_attr_id = 1;
	cb_literal_id = 1;
	cb_field_id = 1;
	cb_stack_size = COB_STACK_SIZE;

#ifdef	_WIN32
	/* Allows running tests under Win */
	p = getenv ("COB_UNIX_LF");
	if (p && (*p == 'Y' || *p == 'y' || *p == '1')) {
		_setmode (_fileno (stdin), _O_BINARY);
		_setmode (_fileno (stdout), _O_BINARY);
		_setmode (_fileno (stderr), _O_BINARY);
	}
#endif

#ifdef	HAVE_SETLOCALE
	setlocale (LC_ALL, "");
	setlocale (LC_NUMERIC, "C");
#endif

#ifdef	ENABLE_NLS
	bindtextdomain (PACKAGE, LOCALEDIR);
	textdomain (PACKAGE);
#endif

#ifndef _WIN32
	cob_process_id = getpid ();
#endif

	/* Initialize variables */

	/* Set up build time stamp */
	memset (month, 0, sizeof(month));
	day = 0;
	year = 0;
	sscanf (__DATE__, "%s %d %d", month, &day, &year);
	if (day && year) {
		snprintf (cobc_buffer, COB_MINI_MAX, "%s %2.2d %4.4d %s",
			  month, day, year, __TIME__);
	} else {
		snprintf (cobc_buffer, COB_MINI_MAX, "%s %s", __DATE__, __TIME__);
	}
	size = strlen (cobc_buffer);
	cb_oc_build_stamp = cobc_malloc (size + 1);
	memcpy (cb_oc_build_stamp, cobc_buffer, size);

	output_name = NULL;

	if ((p = cobc_getenv ("TMPDIR")) != NULL) {
		cob_tmpdir = p;
	} else if ((p = cobc_getenv ("TMP")) != NULL) {
		cob_tmpdir = p;
		p = cobc_malloc (strlen (cob_tmpdir) + 10);
		sprintf (p, "TMPDIR=%s", cob_tmpdir);
		putenv (p);
	} else {
		cob_tmpdir = "/tmp";
		putenv ((char *)"TMPDIR=/tmp");
	}

	cob_cc = cobc_getenv ("COB_CC");
	if (cob_cc == NULL) {
		cob_cc = COB_CC;
	}
	cob_config_dir = cobc_getenv ("COB_CONFIG_DIR");
	if (cob_config_dir == NULL) {
		cob_config_dir = COB_CONFIG_DIR;
	}
	cob_copy_dir = cobc_getenv ("COB_COPY_DIR");
	if (cob_copy_dir == NULL) {
		cob_copy_dir = COB_COPY_DIR;
	}

	p = cobc_getenv ("COB_CFLAGS");
	if (p) {
		COBC_ADD_STR (cob_cflags, p, NULL, NULL);
	} else {
		COBC_ADD_STR (cob_cflags, COB_CFLAGS, NULL, NULL);
	}

	p = cobc_getenv ("COB_LDFLAGS");
	if (p) {
		COBC_ADD_STR (cob_ldflags, p, NULL, NULL);
	} else {
		COBC_ADD_STR (cob_ldflags, COB_LDFLAGS, NULL, NULL);
	}

	p = cobc_getenv ("COB_LIBS");
	if (p) {
		COBC_ADD_STR (cob_libs, p, NULL, NULL);
	} else {
		COBC_ADD_STR (cob_libs, COB_LIBS, NULL, NULL);
	}

	p = cobc_getenv ("COB_LDADD");
	if (p) {
		COBC_ADD_STR (cob_libs, " ", p, NULL);
	}

	p = cobc_getenv ("COB_LIB_PATHS");
	if (p) {
		COBC_ADD_STR (cob_lib_paths, p, NULL, NULL);
	} else {
		COBC_ADD_STR (cob_lib_paths, " ", NULL, NULL);
	}

	p = getenv ("COB_EBCDIC");
	if (p && (*p == 'F' || *p == 'f')) {
		alt_ebcdic = 1;
	}

	/* Set up abort callback */
	if (setjmp (cobc_jmpbuf) != 0) {
		if (cb_source_file) {
			fprintf (stderr, _("Aborting compile of %s at line %d\n"),
				 cb_source_file, cb_source_line);
		}
		fflush (stderr);
		if (yyout) {
			fflush (yyout);
		}
		if (cb_storage_file) {
			fflush (cb_storage_file);
		}
		cobc_clean_up (99);
		return 99;
	}

	/* Set default computed goto usage if appropriate */
#ifdef	__GNUC__
	cb_flag_computed_goto = 1;
#elif	defined(__SUNPRO_C) && __SUNPRO_C >= 0x570
	cb_flag_computed_goto = 1;
#elif	defined(__xlc__) && defined(__IBMC__) && __IBMC__ >= 700
	cb_flag_computed_goto = 1;
	COBC_ADD_STR (cob_cflags, " -qlanglvl=extended",
		      NULL, NULL);
#endif

	/* Process command line arguments */
	iargs = process_command_line (argc, argv);

	/* Check the filename */
	if (iargs == argc) {
		fprintf (stderr, _("cobc: No input files\n"));
		exit (1);
	}

	/* Compiler special options */

#ifdef	_MSC_VER
	if (!verbose_output) {
		COBC_ADD_STR (cob_cflags, " /nologo", NULL, NULL);
	}
#if	_MSC_VER >= 1400
	if (!verbose_output) {
		manicmd = "mt /nologo";
		manilink = "/link /nologo /manifest";
	} else {
		manicmd = "mt";
		manilink = "/link /manifest";
	}
#else
	manilink = "/link";
#endif
#endif

#ifdef	__INTEL_COMPILER
	COBC_ADD_STR (cob_cflags, " -vec-report0 -opt-report 0",
		      NULL, NULL);
#elif	defined(__GNUC__)
	COBC_ADD_STR (cob_cflags, " -Wno-unused -fsigned-char -fno-gcse",
		      NULL, NULL);
#ifdef	HAVE_PSIGN_OPT
	COBC_ADD_STR (cob_cflags, " -Wno-pointer-sign",
		      NULL, NULL);
#endif
#elif defined(__xlc__)
	COBC_ADD_STR (cob_cflags, " -qro -qroconst",
		      NULL, NULL);
#endif

	/* Process COBCPY environment variable */
	process_env_copy_path ();

	/* Add COB_COPY_DIR directories */
	cb_include_list = cb_text_list_add (cb_include_list, cob_copy_dir);

	/* Defaults are set here */
	if (!cb_flag_syntax_only) {
		if (!wants_nonfinal) {
			if (cobc_flag_main) {
				cb_compile_level = CB_LEVEL_EXECUTABLE;
			}
			if (cobc_flag_module) {
				cb_compile_level = CB_LEVEL_MODULE;
			}
			if (cobc_flag_library) {
				cb_compile_level = CB_LEVEL_LIBRARY;
			}
		}
		if (cb_compile_level == 0 && !wants_nonfinal) {
			cb_compile_level = CB_LEVEL_MODULE;
			cobc_flag_module = 1;
		}
		if (wants_nonfinal && cb_compile_level != CB_LEVEL_PREPROCESS &&
		    !cobc_flag_main && !cobc_flag_module && !cobc_flag_library) {
			cobc_flag_module = 1;
		}
	} else {
			cb_compile_level = CB_LEVEL_TRANSLATE;
	}

	if (output_name && cb_compile_level < CB_LEVEL_LIBRARY &&
	    (argc - iargs) > 1) {
		fprintf (stderr, _("cobc: -o option invalid in this combination\n"));
		exit (1);
	}

	if (cb_flag_sign_ascii && cb_flag_sign_ebcdic) {
		fprintf (stderr, _("Only one of -fsign-ascii or -fsign-ebcdic may be specified\n"));
		exit (1);
	}

	if (cb_flag_sign_ascii) {
		cb_display_sign = COB_DISPLAY_SIGN_ASCII;
	}
	if (cb_flag_sign_ebcdic) {
		cb_display_sign = COB_DISPLAY_SIGN_EBCDIC;
	}
	if (cb_flag_notrunc) {
		cb_binary_truncate = 0;
		cb_pretty_display = 0;
	}

	/* Do module initialization */
	cobc_init_scanner ();
	cobc_init_reserved ();
	cobc_init_typeck ();

	/* Set defines */
	if (cb_perform_osvs) {
		cb_define_list = cb_define_list_add (cb_define_list,
						     "PERFORM-TYPE='OSVS'");
	} else {
		cb_define_list = cb_define_list_add (cb_define_list,
						     "PERFORM-TYPE='MF'");
	}
	if (cb_flag_sign_ebcdic) {
		cb_define_list = cb_define_list_add (cb_define_list,
						     "SIGN='EBCDIC'");
	} else {
		cb_define_list = cb_define_list_add (cb_define_list,
						     "SIGN='ASCII'");
	}

	/* If -P=file specified, all lists go to this file */
	if (cobc_list_file) {
		cb_listing_file = fopen (cobc_list_file, "w");
		if (!cb_listing_file) {
			cobc_terminate (cobc_list_file);
		}
	}

	/* Process input files */
	status = 0;
	while (iargs < argc) {
		/* Set up file parameters */
		fn = process_filename (argv[iargs++]);
		if (!fn) {
			cobc_clean_up (1);
			return 1;
		}

		/* Preprocess */
		if (cb_compile_level >= CB_LEVEL_PREPROCESS &&
		    fn->need_preprocess) {
			status |= preprocess (fn);
		}
		if (cobc_list_file) {
			putc ('\n', cb_listing_file);
		}
	}

	if (cobc_list_file) {
		fclose (cb_listing_file);
		cb_listing_file = NULL;
	}

	if (status) {
		cobc_clean_up (1);
		return 1;
	}

	for (fn = file_list; fn; fn = fn->next) {
		cb_id = 1;
		cb_attr_id = 1;
		cb_literal_id = 1;
		cb_field_id = 1;
		iparams++;
		demangle_name = fn->demangle_source;
		memset (optimize_defs, 0, sizeof(optimize_defs));
		if (iparams > 1 && cb_compile_level == CB_LEVEL_EXECUTABLE &&
		    !cb_flag_syntax_only) {
			local_level = cb_compile_level;
			cobc_flag_main = 0;
			cb_compile_level = CB_LEVEL_ASSEMBLE;
		}
		/* Translate */
		if (cb_compile_level >= CB_LEVEL_TRANSLATE &&
		    fn->need_translate) {
			if (process_translate (fn) != 0) {
				cobc_clean_up (1);
				return 1;
			}
		}
		if (cb_flag_syntax_only) {
			continue;
		}

		/* Compile */
		if (cb_compile_level == CB_LEVEL_COMPILE) {
			if (process_compile (fn) != 0) {
				cobc_clean_up (1);
				return 1;
			}
		}

		/* Build module */
		if (cb_compile_level == CB_LEVEL_MODULE && fn->need_assemble) {
			if (process_module_direct (fn) != 0) {
				cobc_clean_up (1);
				return 1;
			}
		} else {
			/* Assemble */
			if (cb_compile_level >= CB_LEVEL_ASSEMBLE &&
			    fn->need_assemble) {
				if (process_assemble (fn) != 0) {
					cobc_clean_up (1);
					return 1;
				}
			}

			/* Build module */
			if (cb_compile_level == CB_LEVEL_MODULE) {
				if (process_module (fn) != 0) {
					cobc_clean_up (1);
					return 1;
				}
			}
		}
	}

	if (!cb_flag_syntax_only) {
		/* Link */
		if (local_level == CB_LEVEL_EXECUTABLE) {
			cb_compile_level = CB_LEVEL_EXECUTABLE;
			cobc_flag_main = 1;
		}
		if (cb_compile_level == CB_LEVEL_LIBRARY) {
			/* Multi-program module */
			if (process_library (file_list) != 0) {
				cobc_clean_up (1);
				return 1;
			}
		} else if (cb_compile_level == CB_LEVEL_EXECUTABLE) {
			/* Executable */
			if (process_link (file_list) != 0) {
				cobc_clean_up (1);
				return 1;
			}
		}
	}

	/* We have successfully completed */
	cobc_clean_up (0);

	return 0;
}
