/*
   Copyright (C) 2001,2002,2003,2004,2005,2006,2007 Keisuke Nishida
   Copyright (C) 2007,2008,2009,2010 Roger While

   This file is part of OpenCOBOL.

   The OpenCOBOL compiler is free software: you can redistribute it
   and/or modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.

   OpenCOBOL is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with OpenCOBOL.  If not, see <http://www.gnu.org/licenses/>.
*/


%expect 0

%defines
%error-verbose
%verbose
%name-prefix="pp"

%{
#include "config.h"

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "cobc.h"

#define pperror ppparse_error

#define COND_EQ		0
#define COND_LT		1U
#define COND_GT		2U
#define COND_LE		3U
#define COND_GE		4U
#define COND_NE		5U

/* Local variables */

static unsigned int	current_cmd;

#if	0	/* RXWRXW OPT */
static const char	* const compopts[] = {
	"ibm",
	"ibmcomp",
	"iso2002",
	"mf",
	"mfcomment",
	"sticky-linkage",
	"trunc",
	"noibmcomp",
	"nofold-copy-name",
	"nofoldcopyname",
	"nomfcomment",
	"nosticky-linkage",
	"notrunc"
};

static const char	* const varopts[] = {
	"fold-copy-name",
	"foldcopyname",
	"sourceformat",
	"trunc"
};
#endif

/* Local functions */

static char *
fix_filename (char *name)
{
	/* remove quotation from alphanumeric literals */
	if (name[0] == '\'' || name[0] == '\"') {
		name++;
		name[strlen (name) - 1] = 0;
	}
	return name;
}

static char *
fold_lower (char *name)
{
	unsigned char	*p;

	for (p = (unsigned char *)name; *p; p++) {
		if (isupper (*p)) {
			*p = tolower (*p);
		}
	}
	return name;
}

static char *
fold_upper (char *name)
{
	unsigned char	*p;

	for (p = (unsigned char *)name; *p; p++) {
		if (islower (*p)) {
			*p = toupper (*p);
		}
	}
	return name;
}

static struct cb_replace_list *
cb_replace_list_add (struct cb_replace_list *list,
		     struct cb_text_list *old_text,
		     struct cb_text_list *new_text,
		     size_t lead_or_trail)
{
	struct cb_replace_list *p;
	struct cb_replace_list *l;

	p = cobc_malloc (sizeof (struct cb_replace_list));
	p->next = NULL;
	p->old_text = old_text;
	p->new_text = new_text;
	p->leading_trailing = lead_or_trail;
	if (!list) {
		return p;
	} else {
		for (l = list; l->next; l = l->next) {
			;
		}
		l->next = p;
		return list;
	}
}

static unsigned int
cb_compare_vals (struct cb_define_struct *p1, struct cb_define_struct *p2,
		 unsigned int cond)
{
	int	result;

	if (!p1 || !p2) {
		return 0;
	}
	if (p1->deftype != PLEX_DEF_LIT && p1->deftype != PLEX_DEF_NUM) {
		return 0;
	}
	if (p2->deftype != PLEX_DEF_LIT && p2->deftype != PLEX_DEF_NUM) {
		return 0;
	}
	if (p1->deftype != p2->deftype) {
		ppparse_error (_("Invalid directive comparison"));
		return 0;
	}
	if (p1->deftype == PLEX_DEF_LIT) {
		result = strcmp (p1->value, p2->value);
	} else {
		if (p1->sign && !p2->sign) {
			result = -1;
		} else if (!p1->sign && p2->sign) {
			result = 1;
		} else if (p1->int_part < p2->int_part) {
			if (p1->sign) {
				result = 1;
			} else {
				result = -1;
			}
		} else if (p1->int_part > p2->int_part) {
			if (p1->sign) {
				result = -1;
			} else {
				result = 1;
			}
		} else if (p1->dec_part < p2->dec_part) {
			if (p1->sign) {
				result = 1;
			} else {
				result = -1;
			}
		} else if (p1->dec_part > p2->dec_part) {
			if (p1->sign) {
				result = -1;
			} else {
				result = 1;
			}
		} else {
			result = 0;
		}
	}
	switch (cond) {
	case COND_EQ:
		return (result == 0);
		break;
	case COND_LT:
		return (result < 0);
		break;
	case COND_GT:
		return (result > 0);
		break;
	case COND_LE:
		return (result <= 0);
		break;
	case COND_GE:
		return (result >= 0);
		break;
	case COND_NE:
		return (result != 0);
		break;
	default:
		break;
	}
	return 0;
}

static struct cb_define_struct *
cb_define_add (struct cb_define_struct *list, const char *name,
	       const char *text, const unsigned int override)
{
	struct cb_define_struct	*p;
	struct cb_define_struct	*l;

	/* Check duplicate */
	for (l = list; l; l = l->next) {
		if (!strcasecmp (name, l->name)) {
			if (!override && l->deftype != PLEX_DEF_DEL) {
				ppparse_error (_("Duplicate define"));
				return NULL;
			}
			if (l->value) {
				free (l->value);
				l->value = NULL;
			}
			if (cb_set_value (l, text)) {
				ppparse_error (_("Invalid constant"));
				return NULL;
			}
			return list;
		}
	}

	p = cobc_malloc (sizeof (struct cb_define_struct));
	p->next = NULL;
	p->name = cobc_strdup (name);
	if (cb_set_value (p, text)) {
		ppparse_error (_("Invalid constant"));
		free (p->name);
		free (p);
		return NULL;
	}

	if (!list) {
		return p;
	}
	for (l = list; l->next; l = l->next) {
		;
	}
	l->next = p;
	return list;
}

static void
cb_define_del (const char *name)
{
	struct cb_define_struct	*l;

	for (l = cb_setvar_list; l; l = l->next) {
		if (!strcmp (name, l->name)) {
			l->deftype = PLEX_DEF_DEL;
			if (l->value) {
				free (l->value);
				l->value = NULL;
			}
			l->sign = 0;
			l->int_part = 0;
			l->dec_part = 0;
			break;
		}
	}
}

static struct cb_define_struct *
cb_search_lists (const char *name)
{
	struct cb_define_struct	*p;

	for (p = cb_define_list; p; p = p->next) {
		if (!strcasecmp (name, p->name)) {
			return p;
			break;
		}
	}
	for (p = cb_setvar_list; p; p = p->next) {
		if (!strcasecmp (name, p->name)) {
			if (p->deftype != PLEX_DEF_DEL) {
				return p;
			}
			break;
		}
	}
	return NULL;
}

static unsigned int
ppp_search_comp_vars (const char *name)
{
#undef	CB_PARSE_DEF
#define	CB_PARSE_DEF(x,z)	if (!strcasecmp (name, x)) return (z);
#include "ppparse.def"
#undef	CB_PARSE_DEF
	return 0;
}

%}

%union {
	char			*s;
	struct cb_text_list	*l;
	struct cb_replace_list	*r;
	struct cb_define_struct	*ds;
	unsigned int		ui;
	int			si;
};

%token TOKEN_EOF 0	"end of file"

%token BY
%token COPY
%token EQEQ		"=="
%token IN
%token LEADING
%token OF
%token OFF
%token PRINTING
%token REPLACE
%token REPLACING
%token SUPPRESS
%token TRAILING
%token DOT		"."

%token SOURCE_DIRECTIVE
%token FORMAT
%token IS
%token FIXED
%token FREE

%token DEFINE_DIRECTIVE
%token AS
%token PARAMETER
%token OVERRIDE

%token SET_DIRECTIVE
%token CONSTANT
%token SOURCEFORMAT
%token FOLDCOPYNAME
%token NOFOLDCOPYNAME

%token IF_DIRECTIVE
%token ELSE_DIRECTIVE
%token ENDIF_DIRECTIVE
%token ELIF_DIRECTIVE

%token GE		">="
%token LE		"<="
%token LT		"<"
%token GT		">"
%token EQ		"="
%token NE		"<>"
%token NOT
%token THAN
%token TO
%token OR
%token EQUAL
%token GREATER
%token LESS
%token SET
%token DEFINED

%token TERMINATOR	"end of line"

%token <s> TOKEN	"Identifier or Literal"
%token <s> VARIABLE	"Variable"
%token <s> LITERAL	"Literal"

%type <s>	copy_in
%type <l>	token_list
%type <l>	identifier
%type <l>	subscripts
%type <l>	text_src
%type <l>	text_dst
%type <l>	text_partial_src
%type <l>	text_partial_dst
%type <r>	copy_replacing
%type <r>	replacing_list

%type <ds>	object_id
%type <ui>	_override
%type <ui>	condition_clause
%type <ui>	_not

%%

statement_list:
| statement_list statement
;

statement:
  copy_statement DOT
| replace_statement DOT
| directive TERMINATOR
;

directive:
  SOURCE_DIRECTIVE source_directive
| DEFINE_DIRECTIVE define_directive
| SET_DIRECTIVE set_directive
| IF_DIRECTIVE
  {
	current_cmd = PLEX_ACT_IF;
  }
  if_directive
| ELIF_DIRECTIVE
  {
	current_cmd = PLEX_ACT_ELIF;
  }
  if_directive
| ELSE_DIRECTIVE
  {
	plex_action_directive (PLEX_ACT_ELSE, 0);
  }
| ENDIF_DIRECTIVE
  {
	plex_action_directive (PLEX_ACT_END, 0);
  }
;

set_directive:
  set_choice
| set_directive set_choice
;

set_choice:
  CONSTANT VARIABLE _as LITERAL
  {
	struct cb_define_struct	*p;

	p = cb_define_add (cb_setvar_list, $2, $4, 1);
	if (p) {
		cb_setvar_list = p;
	}
	fprintf (ppout, "#deflit %s %s\n", $2, $4);
  }
| VARIABLE set_options
| SOURCEFORMAT _as LITERAL
  {
	char	*p;
	size_t	size;
	int	quote;

	p = $3;
	if (*p == '\"' || *p == '\'') {
		quote = *p;
		p++;
		size = strlen (p) - 1;
		p[size] = 0;
	}
	if (!strcasecmp (p, "FIXED")) {
		cb_source_format = CB_FORMAT_FIXED;
	} else if (!strcasecmp (p, "FREE")) {
		cb_source_format = CB_FORMAT_FREE;
	} else {
		ppparse_error (_("Invalid SOURCEFORMAT directive"));
	}
  }
| NOFOLDCOPYNAME
  {
	cb_flag_fold_copy_upper = 0;
	cb_flag_fold_copy_lower = 0;
  }
| FOLDCOPYNAME _as LITERAL
  {
	char	*p;
	size_t	size;
	int	quote;

	p = $3;
	if (*p == '\"' || *p == '\'') {
		quote = *p;
		p++;
		size = strlen (p) - 1;
		p[size] = 0;
	}
	if (!strcasecmp (p, "UPPER")) {
		cb_flag_fold_copy_upper = 1;
		cb_flag_fold_copy_lower = 0;
	} else if (!strcasecmp (p, "LOWER")) {
		cb_flag_fold_copy_upper = 0;
		cb_flag_fold_copy_lower = 1;
	} else {
		ppparse_error (_("Invalid FOLD-COPY-NAME directive"));
	}
  }
;

set_options:
  /* empty */
  {
	fprintf (ppout, "#option %s\n", $<s>0);
  }
| _as LITERAL
  {
	fprintf (ppout, "#option %s %s\n", $<s>0, $2);
  }
;

source_directive:
  _format _is format_type
;

format_type:
  FIXED
  {
	cb_source_format = CB_FORMAT_FIXED;
  }
| FREE
  {
	cb_source_format = CB_FORMAT_FREE;
  }
;

define_directive:
  VARIABLE _as OFF
  {
	cb_define_del ($1);
  }
| VARIABLE _as PARAMETER _override
  {
	char			*s;
	struct cb_define_struct	*p;

	s = getenv ($1);
	if (s && *s && *s != ' ') {
		p = cb_define_add (cb_setvar_list, $1, s, $4);
		if (p) {
			cb_setvar_list = p;
		}
	}
  }
| VARIABLE _as LITERAL _override
  {
	struct cb_define_struct	*p;

	p = cb_define_add (cb_setvar_list, $1, $3, $4);
	if (p) {
		cb_setvar_list = p;
	}
  }
| CONSTANT VARIABLE _as LITERAL _override
  {
	struct cb_define_struct	*p;

	p = cb_define_add (cb_setvar_list, $2, $4, $5);
	if (p) {
		cb_setvar_list = p;
	}
	fprintf (ppout, "#deflit %s %s\n", $2, $4);
  }
;


if_directive:
  VARIABLE _is _not DEFINED
  {
	unsigned int		found;

	found = (cb_search_lists ($1) != NULL);
	plex_action_directive (current_cmd, found ^ $3);
  }
| VARIABLE _is _not SET
  {
	unsigned int		found;

	found = ppp_search_comp_vars ($1);
	plex_action_directive (current_cmd, found ^ $3);
  }
| VARIABLE _is _not condition_clause object_id
  {
	struct cb_define_struct	*p;
	unsigned int		found;

	found = 0;
	p = cb_search_lists ($1);
	found = cb_compare_vals (p, $5, $4);
	plex_action_directive (current_cmd, found ^ $3);
  }
| LITERAL _is _not condition_clause object_id
  {
	struct cb_define_struct	*p;
	unsigned int		found;

	found = 0;
	p = cobc_malloc (sizeof (struct cb_define_struct));
	p->next = NULL;
	if (cb_set_value (p, $1)) {
		ppparse_error (_("Invalid constant"));
		free (p);
	} else {
		found = cb_compare_vals (p, $5, $4);
	}
	plex_action_directive (current_cmd, found ^ $3);
  }
;

object_id:
  LITERAL
  {
	struct cb_define_struct	*p;

	p = cobc_malloc (sizeof (struct cb_define_struct));
	p->next = NULL;
	if (cb_set_value (p, $1)) {
		ppparse_error (_("Invalid constant"));
		free (p);
		$$ = NULL;
	} else {
		$$ = p;
	}
  }
| VARIABLE
  {
	struct cb_define_struct	*p;

	p = cb_search_lists ($1);
	if (p != NULL && p->deftype != PLEX_DEF_NONE) {
		$$ = p;
	} else {
		$$ = NULL;
	}
  }
;

condition_clause:
  GREATER _than OR EQUAL _to
  {
	$$ = COND_GE;
  }
| GREATER _than
  {
	$$ = COND_GT;
  }
| LESS _than OR EQUAL _to
  {
	$$ = COND_LE;
  }
| LESS _than
  {
	$$ = COND_LT;
  }
| EQUAL _to
  {
	$$ = COND_EQ;
  }
| GE
  {
	$$ = COND_GE;
  }
| GT
  {
	$$ = COND_GT;
  }
| LE
  {
	$$ = COND_LE;
  }
| LT
  {
	$$ = COND_LT;
  }
| EQ
  {
	$$ = COND_EQ;
  }
| NE
  {
	$$ = COND_NE;
  }
;

copy_statement:
  COPY TOKEN copy_in copy_suppress copy_replacing
  {
	fputc ('\n', ppout);
	$2 = fix_filename ($2);
	if (cb_flag_fold_copy_lower) {
		$2 = fold_lower ($2);
	} else if (cb_flag_fold_copy_upper) {
		$2 = fold_upper ($2);
	}
	if ($3) {
		$3 = fix_filename ($3);
		if (cb_flag_fold_copy_lower) {
			$3 = fold_lower ($3);
		} else if (cb_flag_fold_copy_upper) {
			$3 = fold_upper ($3);
		}
	}
	ppcopy ($2, $3, $5);
  }
;

copy_in:
  /* nothing */
  {
	$$ = NULL;
  }
| IN TOKEN
  {
	$$ = $2;
  }
| OF TOKEN
  {
	$$ = $2;
  }
;

copy_suppress:
| SUPPRESS _printing
;

copy_replacing:
  /* nothing */
  {
	$$ = NULL;
  }
| REPLACING replacing_list
  {
	$$ = $2;
  }
;

replace_statement:
  REPLACE replacing_list
  {
	pp_set_replace_list ($2);
  }
| REPLACE OFF
  {
	pp_set_replace_list (NULL);
  }
;

replacing_list:
  text_src BY text_dst
  {
	$$ = cb_replace_list_add (NULL, $1, $3, 0);
  }
| LEADING text_partial_src BY text_partial_dst
  {
	$$ = cb_replace_list_add (NULL, $2, $4, CB_REPLACE_LEADING);
  }
| TRAILING text_partial_src BY text_partial_dst
  {
	$$ = cb_replace_list_add (NULL, $2, $4, CB_REPLACE_TRAILING);
  }
| replacing_list text_src BY text_dst
  {
	$$ = cb_replace_list_add ($1, $2, $4, 0);
  }
| replacing_list LEADING text_partial_src BY text_partial_dst
  {
	$$ = cb_replace_list_add ($1, $3, $5, CB_REPLACE_LEADING);
  }
| replacing_list TRAILING text_partial_src BY text_partial_dst
  {
	$$ = cb_replace_list_add ($1, $3, $5, CB_REPLACE_TRAILING);
  }
;

text_src:
  EQEQ token_list EQEQ
  {
	$$ = $2;
  }
| identifier
  {
	$$ = $1;
  }
;

text_dst:
  EQEQ EQEQ
  {
	$$ = NULL;
  }
| EQEQ token_list EQEQ
  {
	$$ = $2;
  }
| identifier
  {
	$$ = $1;
  }
;

text_partial_src:
  EQEQ TOKEN EQEQ
  {
	$$ = cb_text_list_add (NULL, $2);
  }
;

text_partial_dst:
  EQEQ EQEQ
  {
	$$ = NULL;
  }
| EQEQ TOKEN EQEQ
  {
	$$ = cb_text_list_add (NULL, $2);
  }
;

token_list:
  TOKEN
  {
	$$ = cb_text_list_add (NULL, $1);
  }
| token_list TOKEN
  {
	$$ = cb_text_list_add ($1, $2);
  }
;

identifier:
  TOKEN
  {
	$$ = cb_text_list_add (NULL, $1);
  }
| identifier IN TOKEN
  {
	$$ = cb_text_list_add ($1, " ");
	$$ = cb_text_list_add ($$, "IN");
	$$ = cb_text_list_add ($$, " ");
	$$ = cb_text_list_add ($$, $3);
  }
| identifier OF TOKEN
  {
	$$ = cb_text_list_add ($1, " ");
	$$ = cb_text_list_add ($$, "OF");
	$$ = cb_text_list_add ($$, " ");
	$$ = cb_text_list_add ($$, $3);
  }
| identifier '(' subscripts ')'
  {
	struct cb_text_list *l;

	$$ = cb_text_list_add ($1, " ");
	$$ = cb_text_list_add ($$, "(");
	$3 = cb_text_list_add ($3, ")");
	for (l = $$; l->next; l = l->next) {
		;
	}
	l->next = $3;
  }
;

subscripts:
  TOKEN
  {
	$$ = cb_text_list_add (NULL, $1);
  }
| subscripts TOKEN
  {
	$$ = cb_text_list_add ($1, " ");
	$$ = cb_text_list_add ($$, $2);
  }
;

/* Optional keywords */

_override:
  /* empty */
  {
	$$ = 0;
  }
| OVERRIDE
  {
	$$ = 1;
  }
;

_not:
  /* empty */
  {
	$$ = 0;
  }
| NOT
  {
	$$ = 1;
  }
;

_as:		| AS ;
_format:	| FORMAT ;
_is:		| IS ;
_printing:	| PRINTING ;
_than:		| THAN ;
_to:		| TO ;

%%
